# -*- coding: utf-8 -*-
"""
.NET Streams

REFERENCES

    https://www.ntcore.com/files/dotnetformat.htm
    https://referencesource.microsoft.com/System.AddIn/System/Addin/MiniReflection/MetadataReader/Metadata.cs.html#123
    ECMA-335 6th Edition, June 2012, Section II.24.2.4 #US and #Blob heaps

Copyright (c) 2020-2024 MalwareFrank
"""

import struct as _struct
import logging
from typing import Dict, List, Tuple, Union, Optional
from binascii import hexlify as _hexlify

from pefile import MAX_STRING_LENGTH, Structure

from . import base, errors, mdtable
from .utils import read_compressed_int

logger = logging.getLogger(__name__)


class GenericStream(base.ClrStream):
    """
    A generic CLR Stream of unknown type.
    """
    pass


class HeapItemString(base.HeapItem):
    """
    A HeapItemString is a HeapItem with an encoding.  The .value member
    is the decoded string or None if there was a UnicodeDecodeError.

    A HeapItemString can be compared directly to a str.
    """
    encoding: Optional[str]

    def __init__(self, data: bytes, rva: Optional[int] = None, encoding="utf-8"):
        super().__init__(data, rva=rva)
        self.encoding = encoding
        try:
            self.value: Optional[str] = self.__data__.decode(encoding)
        except UnicodeDecodeError as e:
            self.value = None

    def __str__(self) -> str:
        return self.value or ""

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        return super().__eq__(other)


class HeapItemBinary(base.HeapItem):
    """
    A HeapItemBinary is a HeapItem with an item_size.  The .item_size
    is the parsed compressed integer at the RVA in the binary heap.
    The .value is the bytes following the compressed integer.

    A HeapItemBinary can be compared directly to a bytes object.
    """
    item_size: base.CompressedInt

    def __init__(self, data: bytes, rva: Optional[int] = None):
        self.rva = rva
        # read compressed int, which has a max size of four bytes
        size = base.CompressedInt.read(data[:4], rva)
        if size is None:
            raise ValueError("invalid compressed int")
        self.item_size = size
        base.HeapItem.__init__(self, data[:self.item_size.raw_size + self.item_size], rva)

        # read data
        offset = self.item_size.raw_size
        self.value = self.__data__[offset:offset + self.item_size]

    def value_bytes(self):
        return self.__data__[self.item_size.raw_size:]

    def __eq__(self, other):
        if isinstance(other, base.HeapItem):
            return base.HeapItem.__eq__(self, other)
        if isinstance(other, bytes):
            return self.value == other
        return False


class StringsHeap(base.ClrHeap):
    offset_size = 0

    def get_str(self, index, max_length=MAX_STRING_LENGTH, encoding="utf-8", as_bytes=False):
        """
        Given an index (offset), read a null-terminated UTF-8 (or given encoding) string.
        Returns None on error, or string, or bytes if as_bytes is True.
        """

        item = self.get(index, max_length, encoding)

        if item is None:
            return None

        if as_bytes:
            return item.value_bytes()

        return item.value

    def get(self, index, max_length=MAX_STRING_LENGTH, encoding="utf-8") -> Optional[HeapItemString]:
        """
        Given an index (offset), read a null-terminated UTF-8 (or given encoding) string.
        Returns a HeapItemString, or None on error.
        """
        if not self.__data__ or index is None or not max_length:
            return None

        if index >= len(self.__data__):
            raise IndexError("index out of range")

        offset = index
        end = self.__data__.find(b"\x00", offset)
        if end - offset > max_length:
            return None

        item = HeapItemString(self.__data__[offset:end], rva=self.rva + offset, encoding=encoding)

        return item


class BinaryHeap(base.ClrHeap):
    def get_with_size(self, index: int) -> Optional[Tuple[bytes, int]]:
        try:
            item = self.get(index)
        except IndexError:
            return None

        if item is None:
            return None

        return item.value_bytes(), item.raw_size

    def get_bytes(self, index: int) -> Optional[bytes]:
        try:
            item = self.get(index)
        except IndexError:
            return None

        if item is None:
            return None

        return item.value_bytes()

    def get(self, index: int) -> Optional[HeapItemBinary]:
        if self.__data__ is None:
            logger.warning("stream has no data")
            return None

        if index >= len(self.__data__):
            logger.warning("stream is too small: wanted: 0x%x found: 0x%x", index, len(self.__data__))
            return None

        offset = index

        try:
            item = HeapItemBinary(self.__data__[index:], rva=self.rva + offset)
        except ValueError as e:
            # possible invalid compressed int length, such as invalid leading flags.
            logger.warning(f"stream entry error - {e} @ RVA=0x{hex(self.rva + offset)}")
            return None

        return item


class BlobHeap(BinaryHeap):
    pass


class UserString(HeapItemBinary, HeapItemString):
    """
    The #US or UserStrings stream should contain UTF-16 strings.
    Each entry in the stream includes a byte indicating whether
    any Unicode characters require handling beyond that normally
    provided for 8-bit encoding sets.

    Reference ECMA-335, Partition II Section 24.2.4
    """

    flag: Optional[int] = None

    def __init__(self, data: Union[bytes, HeapItemBinary], rva: Optional[int] = None, encoding="utf-16"):
        self.encoding = encoding
        if isinstance(data, bytes):
            HeapItemBinary.__init__(self, data, rva=rva)
        elif isinstance(data, HeapItemBinary):
            HeapItemBinary.__init__(self, data.raw_data, rva=rva or data.rva)

        buf = self.__data__[self.item_size.raw_size:]
        if self.item_size % 2 == 1:
            # > This final byte holds the value 1 if and only if any UTF16
            # > character within the string has any bit set in its top byte,
            # > or its low byte is any of the following:
            # > 0x01–0x08, 0x0E–0x1F, 0x27, 0x2D, 0x7F.
            #
            # via ECMA-335 6th edition, II.24.2.4
            #
            # Trim this trailing flag, which is not part of the string.
            self.flag = buf[-1]
            str_buf = buf[:-1]
            if self.flag == 0x00:
                # > Otherwise, it holds 0.
                #
                # via ECMA-335 6th edition, II.24.2.4
                #
                # *should* be a normal UTF-16 string, but still not
                # make sense.
                pass
            elif self.flag == 0x01:
                # > The 1 signifies Unicode characters that require handling
                # > beyond that normally provided for 8-bit encoding sets.
                #
                # via ECMA-335 6th edition, II.24.2.4
                #
                # these strings are probably best interpreted as bytes.
                pass
            else:
                logger.warning(f"unexpected string flag value: 0x{self.flag:02x}")
        else:
            logger.warning("string missing trailing flag")
            str_buf = buf

        try:
            self.value = str_buf.decode(encoding)
        except UnicodeDecodeError as e:
            logger.warning(f"UserString decode error (rva:0x{self.rva:08x}): {e}")
            self.value = None

    def value_bytes(self):
        if self.flag is None:
            return self.__data__[self.item_size.raw_size:]
        return self.__data__[self.item_size.raw_size:-1]

    def __eq__(self, other):
        return HeapItemString.__eq__(self, other)


class UserStringHeap(BinaryHeap):
    def get_bytes(self, index) -> Optional[bytes]:
        item = self.get(index)

        if item is None:
            return None

        return item.value_bytes()

    def get(self, index, encoding="utf-16") -> Optional[UserString]:
        bin_item = super().get(index)
        if bin_item is None:
            return None

        us_item = UserString(bin_item, encoding=encoding)

        return us_item


class HeapItemGuid(base.HeapItem):

    def __init__(self, data: bytes, rva: Optional[int] = None):
        super().__init__(data, rva)

    @property
    def value(self):
        return self.__data__

    def __str__(self):
        data = self.__data__
        parts = _struct.unpack_from("<IHH", data)
        part3 = _hexlify(data[8:10])
        part4 = _hexlify(data[10:16])
        part3 = part3.decode("ascii")
        part4 = part4.decode("ascii")
        return f"{parts[0]:08x}-{parts[1]:04x}-{parts[2]:04x}-{part3}-{part4}"

    def __repr__(self):
        return f"HeapItemGuid(data={self.__data__},rva={self.rva})"


class GuidHeap(base.ClrHeap):
    offset_size = 0

    def get_str(self, index, as_bytes=False):
        item = self.get(index)

        if item is None:
            return None

        if as_bytes:
            return item.value_bytes()

        return str(item)

    def get(self, index) -> Optional[HeapItemGuid]:
        if index is None or index < 1:
            return None

        size = 128 // 8  # number of bytes in a guid
        # offset into the GUID stream
        offset = (index - 1) * size

        if offset + size > len(self.__data__):
            raise IndexError("index out of range")

        item = HeapItemGuid(self.__data__[offset:offset + size], self.rva + offset)

        return item


class MDTablesStruct(Structure):
    Reserved_1: int
    MajorVersion: int
    MinorVersion: int
    HeapOffsetSizes: int
    Reserved_2: int
    MaskValid: int
    MaskSorted: int


class MetaDataTables(base.ClrStream):
    """Holds CLR (.NET) Metadata Tables.

    struct:     the stream list entry
    header:     IMAGE_CLR_METADATA_TABLES structure
    tables:     dict of tables where table number is key and value is CLRMetaDataTable object
    tables_list:            list of tables, in processing order
    strings_offset_size:    number of bytes
    guids_offset_size:      number of bytes
    blobs_offset_size:      number of bytes
    """

    _format = (
        "IMAGE_CLR_METADATA_TABLES",
        (
            "I,Reserved_1",
            "B,MajorVersion",
            "B,MinorVersion",
            "B,HeapOffsetSizes",
            "B,Reserved_2",
            "Q,MaskValid",
            "Q,MaskSorted",
        ),
    )

    header: Optional[MDTablesStruct]
    tables: Dict[Union[str, int], base.ClrMetaDataTable]
    tables_list: List[base.ClrMetaDataTable]
    strings_offset_size: int
    guids_offset_size: int
    blobs_offset_size: int

    # from https://www.ntcore.com/files/dotnetformat.htm
    # and https://referencesource.microsoft.com/System.AddIn/System/Addin/MiniReflection/MetadataReader/Metadata.cs.html#123
    Module:                 Optional[mdtable.Module]
    TypeRef:                Optional[mdtable.TypeRef]
    TypeDef:                Optional[mdtable.TypeDef]
    Field:                  Optional[mdtable.Field]
    MethodDef:              Optional[mdtable.MethodDef]
    Param:                  Optional[mdtable.Param]
    InterfaceImpl:          Optional[mdtable.InterfaceImpl]
    MemberRef:              Optional[mdtable.MemberRef]
    Constant:               Optional[mdtable.Constant]
    CustomAttribute:        Optional[mdtable.CustomAttribute]
    FieldMarshal:           Optional[mdtable.FieldMarshal]
    DeclSecurity:           Optional[mdtable.DeclSecurity]
    ClassLayout:            Optional[mdtable.ClassLayout]
    FieldLayout:            Optional[mdtable.FieldLayout]
    StandAloneSig:          Optional[mdtable.StandAloneSig]
    EventMap:               Optional[mdtable.EventMap]
    Event:                  Optional[mdtable.Event]
    PropertyMap:            Optional[mdtable.PropertyMap]
    Property:               Optional[mdtable.Property]
    MethodSemantics:        Optional[mdtable.MethodSemantics]
    MethodImpl:             Optional[mdtable.MethodImpl]
    ModuleRef:              Optional[mdtable.ModuleRef]
    TypeSpec:               Optional[mdtable.TypeSpec]
    ImplMap:                Optional[mdtable.ImplMap]
    FieldRva:               Optional[mdtable.FieldRva]
    Assembly:               Optional[mdtable.Assembly]
    AssemblyProcessor:      Optional[mdtable.AssemblyProcessor]
    AssemblyOS:             Optional[mdtable.AssemblyOS]
    AssemblyRef:            Optional[mdtable.AssemblyRef]
    AssemblyRefProcessor:   Optional[mdtable.AssemblyRefProcessor]
    AssemblyRefOS:          Optional[mdtable.AssemblyRefOS]
    File:                   Optional[mdtable.File]
    ExportedType:           Optional[mdtable.ExportedType]
    ManifestResource:       Optional[mdtable.ManifestResource]
    NestedClass:            Optional[mdtable.NestedClass]
    GenericParam:           Optional[mdtable.GenericParam]
    GenericParamConstraint: Optional[mdtable.GenericParamConstraint]
    Unused:                 Optional[mdtable.Unused]

    def __init__(self, metadata_rva: int, stream_struct: base.StreamStruct, stream_data: bytes):
        super().__init__(metadata_rva, stream_struct, stream_data)
        self.header = None
        self.tables = dict()
        self.tables_list = list()
        strings_offset_size = 0
        guids_offset_size = 0
        blobs_offset_size = 0
        self.Module: Optional[mdtable.Module] = None
        self.TypeRef: Optional[mdtable.TypeRef] = None
        self.TypeDef: Optional[mdtable.TypeDef] = None
        self.Field: Optional[mdtable.Field] = None
        self.MethodDef: Optional[mdtable.MethodDef] = None
        self.Param: Optional[mdtable.Param] = None
        self.InterfaceImpl: Optional[mdtable.InterfaceImpl] = None
        self.MemberRef: Optional[mdtable.MemberRef] = None
        self.Constant: Optional[mdtable.Constant] = None
        self.CustomAttribute: Optional[mdtable.CustomAttribute] = None
        self.FieldMarshal: Optional[mdtable.FieldMarshal] = None
        self.DeclSecurity: Optional[mdtable.DeclSecurity] = None
        self.ClassLayout: Optional[mdtable.ClassLayout] = None
        self.FieldLayout: Optional[mdtable.FieldLayout] = None
        self.StandAloneSig: Optional[mdtable.StandAloneSig] = None
        self.EventMap: Optional[mdtable.EventMap] = None
        self.Event: Optional[mdtable.Event] = None
        self.PropertyMap: Optional[mdtable.PropertyMap] = None
        self.Property: Optional[mdtable.Property] = None
        self.MethodSemantics: Optional[mdtable.MethodSemantics] = None
        self.MethodImpl: Optional[mdtable.MethodImpl] = None
        self.ModuleRef: Optional[mdtable.ModuleRef] = None
        self.TypeSpec: Optional[mdtable.TypeSpec] = None
        self.ImplMap: Optional[mdtable.ImplMap] = None
        self.FieldRva: Optional[mdtable.FieldRva] = None
        self.Assembly: Optional[mdtable.Assembly] = None
        self.AssemblyProcessor: Optional[mdtable.AssemblyProcessor] = None
        self.AssemblyOS: Optional[mdtable.AssemblyOS] = None
        self.AssemblyRef: Optional[mdtable.AssemblyRef] = None
        self.AssemblyRefProcessor: Optional[mdtable.AssemblyRefProcessor] = None
        self.AssemblyRefOS: Optional[mdtable.AssemblyRefOS] = None
        self.File: Optional[mdtable.File] = None
        self.ExportedType: Optional[mdtable.ExportedType] = None
        self.ManifestResource: Optional[mdtable.ManifestResource] = None
        self.NestedClass: Optional[mdtable.NestedClass] = None
        self.GenericParam: Optional[mdtable.GenericParam] = None
        self.GenericParamConstraint: Optional[mdtable.GenericParamConstraint] = None
        self.Unused: Optional[mdtable.Unused] = None

    def parse(self, streams: List[base.ClrStream], lazy_load=False):
        """
        this may raise an exception if the data cannot be parsed correctly.
        however, `self` may still be partially initialized with *some* data.
        """
        STRINGS_MASK = 0x01
        GUIDS_MASK = 0x02
        BLOBS_MASK = 0x04
        DELTA_ONLY_MASK = 0x20
        EXTRA_DATA_MASK = 0x40
        HAS_DELETE_MASK = 0x80
        MAX_TABLES = 64

        # we may be able to parse some data before reaching corruption.
        # so, we'll keep parsing all we can, which updates `self` in-place,
        # and then raise the first deferred exception captured here.
        deferred_exceptions = list()

        header_len = Structure(self.__class__._format).sizeof()
        if not self.__data__ or len(self.__data__) < header_len:
            logger.warning("unable to read .NET metadata tables")
            raise errors.dnFormatError("Unable to read .NET metadata tables")

        #### parse header
        header_struct = MDTablesStruct(self._format, file_offset=self.file_offset)
        header_struct.__unpack__(self.__data__)
        self.header = header_struct

        #### heaps offsets
        if header_struct.HeapOffsetSizes & STRINGS_MASK:
            strings_offset_size = 4
        else:
            strings_offset_size = 2
        if header_struct.HeapOffsetSizes & GUIDS_MASK:
            guids_offset_size = 4
        else:
            guids_offset_size = 2
        if header_struct.HeapOffsetSizes & BLOBS_MASK:
            blobs_offset_size = 4
        else:
            blobs_offset_size = 2
        self.strings_offset_size = strings_offset_size
        self.guids_offset_size = guids_offset_size
        self.blobs_offset_size = blobs_offset_size

        #### heaps
        strings_heap: Optional[StringsHeap] = None
        guid_heap: Optional[GuidHeap] = None
        blob_heap: Optional[BlobHeap] = None
        for s in streams:
            # Find the last instance of the strings, guid, and blob heaps.
            # If there are multiple instances of a type, dotnet runtime uses last.
            if isinstance(s, StringsHeap):
                strings_heap = s
            elif isinstance(s, GuidHeap):
                guid_heap = s
            elif isinstance(s, BlobHeap):
                blob_heap = s

        #### Parse tables rows list.
        #  It is a variable length array of dwords.  Each dword is
        #  the number of rows in a table.  They are ordered by table
        #  number, smallest first.  Only the tables needed/defined
        #  are listed, thus the variable length and need to parse
        #  the header's MaskValid member.
        cur_rva = self.rva + header_len
        table_rowcounts = []
        # read all row counts
        for i in range(MAX_TABLES):
            # if table bit is set
            if header_struct.MaskValid & 2 ** i != 0:
                # read the row count
                table_rowcounts.append(self.get_dword_at_rva(cur_rva))
                # increment to next dword
                cur_rva += 4
            else:
                table_rowcounts.append(0)

        # consume an extra dword if the extra data bit is set
        if header_struct.HeapOffsetSizes & EXTRA_DATA_MASK == EXTRA_DATA_MASK:
            cur_rva += 4

        # initialize all tables
        for i in range(MAX_TABLES):
            # if table bit is set
            if header_struct.MaskValid & 2 ** i:
                is_sorted = header_struct.MaskSorted & 2 ** i != 0
                try:
                    table = mdtable.ClrMetaDataTableFactory.createTable(
                        i,
                        table_rowcounts,
                        is_sorted,
                        self.strings_offset_size,
                        self.guids_offset_size,
                        self.blobs_offset_size,
                        strings_heap,
                        guid_heap,
                        blob_heap,
                        lazy_load,
                    )
                except errors.dnFormatError as e:
                    table = None
                    deferred_exceptions.append(e)
                    logger.warning(str(e))
                if not table:
                    logger.warning("invalid .NET metadata table list @ %d RVA: 0x%x", i, cur_rva)
                    # Everything up to this point has been saved in the object and is accessible,
                    # but more can be parsed, so we delay raising exception.

                    deferred_exceptions.append(errors.dnFormatError(
                        "Invalid .NET metadata table list @ {} rva:{}".format(
                            i, cur_rva
                        ))
                    )
                    continue
                # table number
                table.number = i
                # add to tables dict
                self.tables[i] = table
                # add to tables list
                self.tables_list.append(table)
                # set member, to allow reference by name
                if table.name:
                    setattr(self, table.name, table)

        if lazy_load:
            self._loaded = False

            def full_loader():
                # Called if a property is accessed that requires data from other tables.
                # This will be called multiple times while parsing tables.
                if not self._loaded:
                    self._loaded = True
                    # parse_rows is redundant for lazy loading
                    for table in self.tables_list:
                        table.parse(self.tables_list)

            # Setup lazy loading for all tables
            for table in self.tables_list:
                if table.row_size > 0 and table.num_rows > 0:
                    table_data = self.get_data_at_rva(
                        cur_rva, table.row_size * table.num_rows
                    )
                    table.setup_lazy_load(cur_rva, table_data, full_loader)
                    table.file_offset = self.get_file_offset(cur_rva)
                    cur_rva += table.row_size * table.num_rows
        else:
            #### parse each table
            # here, cur_rva points to start of table rows
            for table in self.tables_list:
                if table.row_size > 0 and table.num_rows > 0:
                    table_data = self.get_data_at_rva(
                        cur_rva, table.row_size * table.num_rows
                    )
                    table.rva = cur_rva
                    table.file_offset = self.get_file_offset(cur_rva)
                    # parse structures (populates .struct for each row)
                    table.parse_rows(cur_rva, table_data)
                    # move to next set of rows
                    cur_rva += table.row_size * table.num_rows
            #### finalize parsing each table
            # For each row, de-references indexes in the .struct and populates row attributes.
            for table in self.tables_list:
                table.parse(self.tables_list)

        # raise warning/error
        if deferred_exceptions:
            raise deferred_exceptions[0]



# -*- coding: utf-8 -*-
import dnfile.utils


def test_compressed_int():
    assert None is dnfile.utils.read_compressed_int(b"")
    assert None is dnfile.utils.read_compressed_int(None)

    assert 0x7f, 1 == dnfile.utils.read_compressed_int(b"\x7f")
    assert 0x3f8f, 2 == dnfile.utils.read_compressed_int(b"\xbf\x8f")
    assert 0x1eadbeef, 4 == dnfile.utils.read_compressed_int(b"\xde\xad\xbe\xef")


def test_struct_char():
    assert None is dnfile.utils.num_bytes_to_struct_char(42)
    assert "Q" == dnfile.utils.num_bytes_to_struct_char(8)
    assert "Q" == dnfile.utils.num_bytes_to_struct_char(5)
    assert "I" == dnfile.utils.num_bytes_to_struct_char(4)
    assert "I" == dnfile.utils.num_bytes_to_struct_char(3)
    assert "H" == dnfile.utils.num_bytes_to_struct_char(2)
    assert "B" == dnfile.utils.num_bytes_to_struct_char(1)
    assert None is dnfile.utils.num_bytes_to_struct_char(0)



import enum
import struct
import datetime
from typing import Any, Dict, List, Tuple, Union, Optional

from . import base, utils, errors, mdtable

# References:
#   https://ntcore.com/files/manifestres.htm
#   https://github.com/0xd4d/dnlib/tree/master/src/DotNet/Resources
#   https://referencesource.microsoft.com/mscorlib/system/resources/resourcetypecode.cs.html


class ResourceTypeCode(enum.IntEnum):
    Null        = 0
    String      = 1
    Boolean     = 2
    Char        = 3
    Byte        = 4
    SByte       = 5
    Int16       = 6
    UInt16      = 7
    Int32       = 8
    UInt32      = 9
    Int64       = 0xa
    UInt64      = 0xb
    Single      = 0xc
    Double      = 0xd
    Decimal     = 0xe
    DateTime    = 0xf
    TimeSpan    = 0x10

    # Types with a special representation, like byte[] and Stream
    ByteArray   = 0x20
    Stream      = 0x21

    # User types - serialized using the binary formatter.
    StartOfUserTypes = 0x40


ResourceTypeStrings: Dict[int, str] = {
    ResourceTypeCode.Null:      "Null",
    ResourceTypeCode.String:    "System.String",
    ResourceTypeCode.Boolean:   "System.Boolean",
    ResourceTypeCode.Char:      "System.Char",
    ResourceTypeCode.Byte:      "System.Byte",
    ResourceTypeCode.SByte:     "System.SByte",
    ResourceTypeCode.Int16:     "System.Int16",
    ResourceTypeCode.UInt16:    "System.UInt16",
    ResourceTypeCode.Int32:     "System.Int32",
    ResourceTypeCode.UInt32:    "System.UInt32",
    ResourceTypeCode.Int64:     "System.Int64",
    ResourceTypeCode.UInt64:    "System.UInt64",
    ResourceTypeCode.Single:    "System.Single",
    ResourceTypeCode.Double:    "System.Double",
    ResourceTypeCode.Decimal:   "System.Decimal",
    ResourceTypeCode.DateTime:  "System.DateTime",
    ResourceTypeCode.TimeSpan:  "System.TimeSpan",
    ResourceTypeCode.ByteArray: "System.ByteArray",
    ResourceTypeCode.Stream:    "System.Stream",
}


class ResourceTypeFactory(object):

    def read_serialized_data(self, data: bytes, offset: int) -> Tuple[bytes, int]:
        """
        Read a serialized data (compressed int size, followed by data of that many bytes).

        Return the data (bytes) and number of bytes read (size + data).
        """
        x = utils.read_compressed_int(data[offset:offset + 4])
        if x is None:
            raise ValueError("CLR resource error: not enough data at offset")
        size = x[0]
        nbytes = x[1]
        offset += nbytes
        value = data[offset:offset + size]
        nbytes += size
        return value, nbytes

    def read_serialized_string(self, data: bytes, offset: int, encoding="utf-8") -> Tuple[str, int]:
        val, n = self.read_serialized_data(data, offset)
        return val.decode(encoding), n

    def read_rsrc_data_v1(self, data: bytes, offset: int, userTypes: List[bytes], entry: "ResourceEntry"):
        t = int.from_bytes(data[offset:offset + 1], byteorder="little", signed=False)
        entry.struct.Type = t
        offset += 4
        # https://github.com/0xd4d/dnlib/blob/master/src/DotNet/Resources/ResourceReader.cs
        if t == -1:
            # Null
            entry.type_name = "Null"
            entry.data = b""
            # entry.value is already None
        elif t < 0 or t >= len(userTypes):
            # invalid resource type
            # TODO warn/error
            return
        # get type string
        ts = userTypes[t]
        # remove comma postfix
        comma_loc = ts.find(b",")
        if comma_loc > 0:
            ts = ts[:comma_loc]
        try:
            tn = ts.decode("utf-8")
        except UnicodeDecodeError as e:
            # TODO warn/error
            tn = None
        entry.type_name = tn
        if entry.type_name:
            d, v = self.type_str_to_type(entry.type_name, data, offset)
            if d is not None:
                entry.data = d
            if v is not None:
                entry.value = v

    def read_rsrc_data_v2(self, data: bytes, offset: int, userTypes: List[bytes], entry: "ResourceEntry"):
        # dnlib reads four bytes, but it looks like one byte in the files I have
        t = int.from_bytes(data[offset:offset + 1], byteorder="little", signed=True)
        entry.struct.Type = t
        offset += 1
        # check for invalid type code
        if t < 0 or t >= ResourceTypeCode.StartOfUserTypes + len(userTypes):
            # invalid resource type
            # TODO warn/error
            return
        # https://github.com/0xd4d/dnlib/blob/master/src/IO/DataReader.cs
        if t in ResourceTypeStrings:
            entry.type_name = ResourceTypeStrings[t]
            d, v = self.type_str_to_type(entry.type_name, data, offset)
            if d is not None:
                entry.data = d
            if v is not None:
                entry.value = v
        elif t >= ResourceTypeCode.StartOfUserTypes:
            # get type string
            ts = userTypes[t - ResourceTypeCode.StartOfUserTypes]
            # remove comma postfix
            comma_loc = ts.find(b",")
            if comma_loc > 0:
                ts = ts[:comma_loc]
            try:
                tn = ts.decode("utf-8")
            except UnicodeDecodeError as e:
                # TODO warn/error
                tn = None
            entry.type_name = tn
        else:
            # TODO
            pass

    def type_str_to_type(self, type_name: str, data: bytes, offset: int) -> Tuple[Optional[bytes], Optional[Any]]:
        """
        Given a type string, data buffer, and offset into that buffer.
        Return a tuple of the raw data and the deserialized value.
        """
        # switch on type
        # https://github.com/0xd4d/dnlib/blob/master/src/IO/DataReader.cs
        final_bytes: Optional[bytes] = None
        final_value: Optional[Any] = None
        if type_name == "Null":
            final_bytes = b""
        elif type_name == "System.String":
            final_bytes, n = self.read_serialized_data(data, offset)
            try:
                final_value = final_bytes.decode("utf-8")
            except UnicodeDecodeError:
                # TODO warn/error
                pass
        elif type_name == "System.Int32":
            tsize = 4
            final_bytes = data[offset:offset + tsize]
            final_value = int.from_bytes(final_bytes, byteorder="little", signed=False)
        elif type_name == "System.Byte":
            tsize = 1
            final_bytes = data[offset:offset + tsize]
            final_value = final_bytes
        elif type_name == "System.SByte":
            tsize = 1
            final_bytes = data[offset:offset + tsize]
            final_value = final_bytes
        elif type_name == "System.Boolean":
            tsize = 1
            final_bytes = data[offset:offset + tsize]
            final_value = final_bytes[0] != 0
        elif type_name == "System.Char":
            tsize = 2
            final_bytes = data[offset:offset + tsize]
            try:
                final_value = final_bytes.decode("utf-16")
            except UnicodeDecodeError:
                # TODO warn/error
                pass
        elif type_name == "System.Int16":
            tsize = 2
            final_bytes = data[offset:offset + tsize]
            final_value = int.from_bytes(final_bytes, byteorder="little", signed=False)
        elif type_name == "System.Int64":
            tsize = 8
            final_bytes = data[offset:offset + tsize]
            final_value = int.from_bytes(final_bytes, byteorder="little", signed=False)
        elif type_name == "System.UInt16":
            tsize = 2
            final_bytes = data[offset:offset + tsize]
            final_value = int.from_bytes(final_bytes, byteorder="little", signed=False)
        elif type_name == "System.UInt32":
            tsize = 4
            final_bytes = data[offset:offset + tsize]
            final_value = int.from_bytes(final_bytes, byteorder="little", signed=False)
        elif type_name == "System.UInt64":
            tsize = 8
            final_bytes = data[offset:offset + tsize]
            final_value = int.from_bytes(final_bytes, byteorder="little", signed=False)
        elif type_name == "System.Single":
            tsize = 4
            final_bytes = data[offset:offset + tsize]
            final_value = struct.unpack("<f", final_bytes)[0]
        elif type_name == "System.Double":
            tsize = 8
            final_bytes = data[offset:offset + tsize]
            final_value = struct.unpack("<d", final_bytes)[0]
        elif type_name == "System.DateTime":
            tsize = 8
            final_bytes = data[offset:offset + tsize]
            x = struct.unpack("<q", final_bytes)[0]
            # Value is stored in lower 62-bits
            # https://github.com/dotnet/runtime/blob/17c55f1/src/libraries/System.Private.CoreLib/src/System/DateTime.cs#L130-L138
            x = x & ((1 << 62) - 1)
            # https://stackoverflow.com/questions/3169517/python-c-sharp-binary-datetime-encoding
            secs = x / 10.0 ** 7
            delta = datetime.timedelta(seconds=secs)
            try:
                dt = datetime.datetime(1, 1, 1) + delta
                final_value = dt
            except OverflowError:
                # TODO warn/error
                pass
        elif type_name == "System.TimeSpan":
            # TODO return resourceDataFactory.Create(new TimeSpan(reader.ReadInt64()));
            tsize = 8
            final_bytes = data[offset:offset + tsize]
        elif type_name == "System.Decimal":
            # https://referencesource.microsoft.com/mscorlib/system/decimal.cs.html
            sign_mask = 0x80000000
            scale_mask = 0x00FF0000
            tsize = 16
            final_bytes = data[offset:offset + tsize]
            low, med, high, flags = struct.unpack("<IIII", final_bytes)
            final_value = low | med << 8 | high << 16
            scale = scale_mask & flags
            if scale > 0:
                final_value = final_value / 10**scale
            if sign_mask & flags:
                final_value = -final_value
        elif type_name == "System.ByteArray":
            tsize = 4
            dsize = int.from_bytes(data[offset:offset + tsize], byteorder="little", signed=False)
            final_bytes = data[offset:offset + tsize + dsize]
            final_value = data[offset + tsize:offset + tsize + dsize]
        elif type_name == "System.Stream":
            tsize = 4
            dsize = int.from_bytes(data[offset:offset + tsize], byteorder="little", signed=False)
            final_bytes = data[offset:offset + tsize + dsize]
            final_value = data[offset + tsize:offset + tsize + dsize]
        else:
            # TODO
            pass
        return final_bytes, final_value


class ExternalResource(base.ClrResource):
    metadata: base.MDTableRow


class FileResource(ExternalResource):
    metadata: mdtable.FileRow


class AssemblyResource(ExternalResource):
    metadata: mdtable.AssemblyRefRow


class InternalResource(base.ClrResource):
    rva: int
    size: int

    def parse(self):
        if not self.data:
            raise errors.rsrcFormatError("No data")
        # attempt to parse as a ResourceSet
        rs = ResourceSet(self.data, self)
        if rs.valid():
            self.data = rs
            rs.parse()
        # otherwise treat as raw resource


class ResourceEntryStruct(object):
    Type: Optional[int]
    Hash: Optional[int]
    NamePtr: Optional[int]
    DataOffset: Optional[int]
    Name: Optional[bytes]

    def __init__(self):
        self.Type: Optional[int] = None
        self.Hash: Optional[int] = None
        self.NamePtr: Optional[int] = None
        self.DataOffset: Optional[int] = None
        self.Name: Optional[bytes] = None


class ResourceEntry(object):
    struct: ResourceEntryStruct
    type_name: Optional[str]
    value: Optional[Any]
    data: Optional[bytes]
    name: Optional[str]

    def __init__(self):
        self.struct = ResourceEntryStruct()
        self.type_name: Optional[str] = None
        self.value: Optional[Any] = None
        self.data: Optional[bytes] = None
        self.name: Optional[str] = None


class ResourceSetStruct(object):
    Magic: int
    NumberOfReaders: int
    SizeOfReaderTypes: int
    ReaderTypes: bytes
    Version: int
    NumberOfResources: int
    NumberOfResourceTypes: int
    DataSectionOffset: int
    TableOfNames: int


class ResourceSet(object):
    parent: Optional[base.ClrResource]
    struct: Optional[ResourceSetStruct]
    entries: List[ResourceEntry]
    MAGIC: int = 0xbeefcace
    MAGIC_BYTES: bytes = b"\xCE\xCA\xEF\xBE"

    def __init__(self, data: bytes, parent: base.ClrResource):
        self._data = data
        self._min_valid_size = 7 * 4
        self.entries: List[ResourceEntry] = list()
        self.resource_types: List[bytes] = list()
        self.struct: Optional[ResourceSetStruct] = None
        self.parent: Optional[base.ClrResource] = parent

    def valid(self):
        if len(self._data) < self._min_valid_size:
            # not enough data
            return False
        # test magic
        if not self._data.startswith(self.MAGIC_BYTES):
            return False
        # we have enough data, and it starts with the right signature
        return True

    def parse(self):
        # parse initial structure
        self.struct = ResourceSetStruct()
        offset = 0
        self.struct.Magic = struct.unpack_from("<I", self._data, offset)[0]
        offset += 4
        self.struct.NumberOfReaders = struct.unpack_from("<I", self._data, offset)[0]
        offset += 4
        self.struct.SizeOfReaderTypes = struct.unpack_from("<I", self._data, offset)[0]
        offset += 4
        # reader types string
        self.struct.ReaderTypes = self._data[offset:offset + self.struct.SizeOfReaderTypes]
        offset += self.struct.SizeOfReaderTypes
        # version
        self.struct.Version = struct.unpack_from("<I", self._data, offset)[0]
        offset += 4
        # NumberOfResources
        self.struct.NumberOfResources = struct.unpack_from("<I", self._data, offset)[0]
        offset += 4
        # NumberofResourceTypes
        self.struct.NumberOfResourceTypes = struct.unpack_from("<I", self._data, offset)[0]
        offset += 4
        # parse more
        for i in range(self.struct.NumberOfResourceTypes):
            # read string length
            try:
                type_string, size = self.read_serialized_data(offset)
            except ValueError:
                raise errors.rsrcFormatError("CLR ResourceSet error: expected more data for types at '{}' rsrc offset {}".format(self.parent.name, offset))
            # add to list
            self.resource_types.append(type_string)
            # next
            offset += size
        # align to 8-byte boundary
        over = offset % 8
        if over != 0:
            offset += 8 - over
        for i in range(self.struct.NumberOfResources):
            e = ResourceEntry()
            e.struct.Hash = int.from_bytes(self._data[offset:offset + 4], byteorder="little")
            self.entries.append(e)
            # next
            offset += 4
        for e in self.entries:
            e.struct.NamePtr = int.from_bytes(self._data[offset:offset + 4], byteorder="little")
            # next
            offset += 4
        # dword (data section offset)
        self.struct.DataSectionOffset = int.from_bytes(self._data[offset:offset + 4], byteorder="little")
        offset += 4
        # table_of_names = current offset
        self.struct.TableOfNames = offset
        rsrc_factory = ResourceTypeFactory()
        problems = list()
        for e in self.entries:
            offset = self.struct.TableOfNames + e.struct.NamePtr
            try:
                e.struct.Name, size = self.read_serialized_data(offset)
                e.name = e.struct.Name.decode("utf-16")
            except UnicodeDecodeError:
                # entry name is initialized to None, so just ignore
                pass
            except ValueError:
                # further entries may be ok; delay this exception
                problems.append("CLR ResourceSet error: expected more data for entries at '{}' rsrc offset {}".format(self.parent.name, offset))
                continue
            offset += size
            e.struct.DataOffset = int.from_bytes(self._data[offset:offset + 4], byteorder="little")
            if self.struct.DataSectionOffset is None:
                continue
            e_data_offset = self.struct.DataSectionOffset + e.struct.DataOffset
            try:
                if self.struct.Version == 1:
                    rsrc_factory.read_rsrc_data_v1(self._data, e_data_offset, self.resource_types, e)
                else:
                    rsrc_factory.read_rsrc_data_v2(self._data, e_data_offset, self.resource_types, e)
            except ValueError:
                # further entries may be ok; delay this exception
                if self.parent:
                    problems.append("CLR ResourceSet error: expected more data for serialized data at '{}' rsrc offset {}".format(self.parent.name, e_data_offset))
                else:
                    problems.append("CLR ResourceSet error: expected more data for serialized data at unknown rsrc offset {}".format(e_data_offset))
            continue
        for p in problems:
            raise errors.rsrcFormatError(p)

    def read_serialized_data(self, offset: int) -> Tuple[bytes, int]:
        """
        Read a serialized data (compressed int size, followed by data of that many bytes).

        Return the data (bytes) and number of bytes read (size + data).
        """
        x = utils.read_compressed_int(self._data[offset:offset + 4])
        if x is None:
            raise ValueError("CLR resource error: not enough data at offset")
        size = x[0]
        nbytes = x[1]
        offset += nbytes
        value = self._data[offset:offset + size]
        nbytes += size
        return value, nbytes

    def read_serialized_string(self, offset: int, encoding="utf-8") -> Tuple[str, int]:
        val, n = self.read_serialized_data(offset)
        return val.decode(encoding), n



import fixtures

import dnfile


def is_mixed_mode(dn: dnfile.dnPE):
    if dn.net is None or dn.net.mdtables is None:
        # no .NET metdata, must be a native PE.
        # therefore, not mixed-mode.
        return False

    if dn.net.Flags and (not dn.net.Flags.CLR_ILONLY):
        return True

    methods = dn.net.mdtables.MethodDef
    assert methods is not None

    for method in methods:
        if method.ImplFlags.miNative:
            # it has a .NET header but also has a native method,
            # so its a mixed-mode assembly.
            return True

    return False


def test_empty_class():
    path = fixtures.DATA / "mixed-mode" / "EmptyClass" / "bin" / "EmptyClass_x86.exe"

    dn = dnfile.dnPE(path)

    assert is_mixed_mode(dn) is True


def test_empty_class_dll():
    path = fixtures.DATA / "mixed-mode" / "EmptyClassDll" / "bin" / "EmptyClass_x86.dll"

    dn = dnfile.dnPE(path)

    assert is_mixed_mode(dn) is True


def test_module_code():
    path = fixtures.DATA / "mixed-mode" / "ModuleCode" / "bin" / "ModuleCode_x86.exe"

    dn = dnfile.dnPE(path)

    assert is_mixed_mode(dn) is True


def test_ilonly():
    path = fixtures.DATA / "hello-world" / "hello-world.exe"

    dn = dnfile.dnPE(path)

    assert is_mixed_mode(dn) is False



# -*- coding: utf-8 -*-
"""
.NET base classes

Copyright (c) 2020-2024 MalwareFrank
"""
import abc
import enum
import struct as _struct
import logging
import functools as _functools
import itertools as _itertools
from typing import TYPE_CHECKING, Any, Dict, List, Type, Tuple, Union, Generic, TypeVar, Optional, Sequence

from pefile import Structure

from . import enums, errors
from .utils import LazyList as _LazyList
from .utils import read_compressed_int as _read_compressed_int

if TYPE_CHECKING:
    from . import stream


logger = logging.getLogger(__name__)


class CompressedInt(int):
    raw_size: int
    __data__: bytes
    value: int
    rva: Optional[int] = None

    def to_bytes(self):
        return self.__data__

    @classmethod
    def read(cls, data: bytes, rva: Optional[int] = None) -> Optional["CompressedInt"]:
        result = _read_compressed_int(data)
        if result is None:
            return None
        ci = CompressedInt(result[0])
        ci.raw_size = result[1]
        ci.value = result[0]
        ci.__data__ = data[:result[1]]
        ci.rva = rva
        return ci


class StreamStruct(Structure):
    Name: bytes
    Offset: int
    Size: int


class ClrStream(abc.ABC):
    def __init__(
        self,
        metadata_rva: int,
        stream_struct: StreamStruct,
        stream_data: bytes,
    ):
        self.struct: StreamStruct = stream_struct
        self.rva: int = metadata_rva + stream_struct.Offset
        self.file_offset: Optional[int] = None
        self.__data__: bytes = stream_data
        self._stream_table_entry_size = stream_struct.sizeof()
        self._data_size = len(stream_data)

    def parse(self, streams: List, lazy_load: bool = False):
        """
        Parse the stream.

        NOTE: do not call until all streams have been initialized,
              since we may need info from other streams.
        """
        pass

    def stream_table_entry_size(self):
        """
        Returns the number of bytes occupied by this entry in the Streams table list.
        """
        return self._stream_table_entry_size

    def sizeof(self):
        """
        Return the size of this stream, in bytes.
        """
        return self._data_size

    def get_file_offset(self, rva=None):
        """
        Return the file offset of the given RVA within this stream.
        If no RVA given, then return the file offset of this stream.
        """
        if self.file_offset is not None and rva is not None:
            stream_offset = rva - self.rva
            if stream_offset >= 0:
                return self.file_offset + stream_offset
        return self.file_offset

    def get_data_at_offset(self, offset, size):
        if size == 0 or offset >= self.sizeof():
            return b""
        return self.__data__[offset:offset + size]

    def get_data_at_rva(self, rva, size):
        offset = rva - self.rva
        return self.get_data_at_offset(offset, size)

    def get_dword_at_rva(self, rva):
        d = self.get_data_at_rva(rva, 4)
        if len(d) < 4:
            return None
        # Little-endian
        i = _struct.unpack("<I", d)[0]
        return i


class HeapItem(abc.ABC):
    """
    HeapItem is a base class for items retrieved from any of the
    heap streams, for example #Strings, #US, #GUID, and #Blob.

    It can be used to access the raw underlying data, the RVA
    from which it was retrieved, an optional interpreted value,
    and the bytes representation of the value.

    Each heap stream .get() call returns a subclass with these
    and optionally additional members.
    """

    rva: Optional[int] = None
    # original data from file
    __data__: bytes
    # interpreted value
    value: Any = None

    def __init__(self, data: bytes, rva: Optional[int] = None):
        self.rva = rva
        self.__data__ = data

    def value_bytes(self):
        """
        Return the raw bytes underlying the interpreted value.

        For the base HeapItem, this is the same as the raw_data.
        """
        return self.__data__

    @property
    def raw_size(self):
        """
        Number of bytes read from the stream, including any header,
        value, and footer.
        """
        return len(self.__data__)

    @property
    def raw_data(self):
        """
        The bytes read from the stream, including any header,
        value, and footer
        """
        return self.__data__

    def __eq__(self, other):
        """
        Two HeapItems are equal if their raw data is the same or their
        interpreted values are the same and not Noney.

        A HeapItem is equal to a bytes object if the HeapItem's value as bytes
        is equal to the bytes object.
        """
        if isinstance(other, HeapItem):
            return self.raw_data == other.raw_data or (self.value is not None and self.value == other.value)
        elif isinstance(other, bytes):
            return self.value_bytes() == other
        return False


class ClrHeap(ClrStream):
    @abc.abstractmethod
    def get(self, index: int):
        raise NotImplementedError()


class RowStruct(Structure):
    pass


class LoadState(enum.Enum):
    Unloaded = 0
    LazyLoaded = 1
    Loaded = 2


class MDTableRow(abc.ABC):
    """
    This is the base class for Metadata Tables' rows.

    A Metadata Table row is a simple structure that holds the
    fields and values.
    """
    #
    # required properties for subclasses.
    #
    # subclasses must define this property,
    # or __init__ will raise an exception.
    #
    # for example:
    #
    #   class ModuleRow(MDTableRow):
    #       _struct_class = ModuleRowStruct
    #
    _struct_class: Type[RowStruct]

    #
    # optional properties for subclasses.
    #
    # when a subclass defines one of these properties,
    # the given fields will be parsed with the appropriate strategy.
    # these strategy defintions:
    #   - map from underlying raw struct (e.g. RowStruct class)
    #   - map to the high-level property (e.g. MDTableRow subclass)
    #
    # for example:
    #
    #   class ModuleRow(MDTableRow):
    #       Name: str
    #       _struct_class = ModuleRowStruct
    #       _struct_strings = {
    #           "Name_StringIndex": "Name",
    #       }
    #
    # this strategy causes the parser to:
    #   1. parse the raw row using `ModuleRowStruct`
    #   2. fetch field `ModuleRowStruct.Name_StringIndex`
    #   3. resolve it as a string (due to strategy name)
    #   4. assign it to field `ModuleRow.Name`
    #
    # valid strategies are:
    #  - asis: map data as-is, possibly change the field name
    #  - strings: resolve via UserString table
    #  - guids: resolve via GUID table
    #  - blobs: resolve via Blob table
    #  - flags: resolve via given flags
    #  - enums: resolve via given enums
    #  - indexes: resolve via given table name
    #  - lists: resolve many items via given table name
    #  - codedindexes: resolve via candidate list of tables
    _struct_strings: Dict[str, str]
    _struct_guids: Dict[str, str]
    _struct_blobs: Dict[str, str]
    _struct_asis: Dict[str, str]
    _struct_codedindexes: Dict[str, Tuple[str, Type["CodedIndex"]]]  # also CodedIndex subclass
    _struct_indexes: Dict[str, Tuple[str, str]]                      # also Metadata table name
    _struct_flags: Dict[str, Tuple[str, Type[enums.ClrFlags]]]       # also ClrFlags subclass
    _struct_enums: Dict[str, Tuple[str, Type[enum.IntEnum]]]         # also enum.IntEnum subclassA
    _struct_lists: Dict[str, Tuple[str, str]]                        # also Metadata table name

    def __init__(
        self,
        tables_rowcounts: List[Optional[int]],
        strings_offset_size: int,
        guid_offset_size: int,
        blob_offset_size: int,
        strings_heap: Optional["stream.StringsHeap"],
        guid_heap: Optional["stream.GuidHeap"],
        blob_heap: Optional["stream.BlobHeap"],
    ):
        """
        Given the tables' row counts and heap info.

        Initialize the following attributes:
            row_size    The size, in bytes, of one row.  Calculated from
                         tables_rowcounts, heap info, and self._format
            struct      The class used to parse the data.

        tables_rowcounts is indexed by table number.  The value is the row count, if it exists, or None.
        """
        assert hasattr(self.__class__, "_struct_class")

        self._loaded = LoadState.Unloaded

        self._tables_rowcnt = tables_rowcounts
        self._strings: Optional["stream.StringsHeap"] = strings_heap
        self._guids: Optional["stream.GuidHeap"] = guid_heap
        self._blobs: Optional["stream.BlobHeap"] = blob_heap
        self._str_offsz = strings_offset_size
        self._guid_offsz = guid_offset_size
        self._blob_offsz = blob_offset_size
        self._format = self._compute_format()
        self._data: bytes = b""

        # we are cheating here: this isn't technically a RowStruct, but actually a RowStruct subclass.
        # but few users will likely reach in here, so ATM its not worth fully type annotating.
        self.struct: RowStruct = self.__class__._struct_class(format=self._format)
        self.row_size: int = self.struct.sizeof()

    @abc.abstractmethod
    def _compute_format(self) -> Tuple[str, Sequence[str]]:
        """
        Compute the structure format.
        This will be passed to RowStruct instances to calcuate the row size and to parse a row.

        This may raise an exception when the offsets to referenced streams is too large (>4 bytes).
        """
        ...

    def set_data(self, data: bytes, file_offset: Optional[int] = None):
        """
        Parse the data and set struct for this row.

        NOTE that the row is not fully parsed, and attributes not set, until
        parse() is called after all tables have had parse_rows() called on them.
        """
        self._data = data
        self.struct = self.__class__._struct_class(format=self._format, file_offset=file_offset)
        self.struct.__unpack__(data)

    # can be safely parsed without all tables being initialized
    CLASS_ATTRS = (
        "_struct_asis", "_struct_strings", "_struct_guids",
        "_struct_blobs", "_struct_flags", "_struct_enums",

    )
    # cannot be fully parsed without all tables being initialized
    CLASS_ATTRS_TABLES = (
        "_struct_codedindexes", "_struct_indexes", "_struct_lists"
    )

    def setup_lazy_load(self, full_loader):
        """Mark this row for lazy-loading.

        `full_loader` will be called if a property is requested that requires
        loading all mdtables before it can be parsed.
        """
        if not self._loaded == LoadState.Unloaded:
            return
        self._loaded = LoadState.LazyLoaded
        self._full_loader = full_loader
        # Retrieve the properties that this row *could* have, along with
        # their associated struct type. These are used to determine what
        # _parse function needs to be called to load the property.
        self._class_struct_attrs, self._class_struct_attrs_tables = (
            _row_class_struct_attrs(self.__class__)
        )

    def __getattr__(self, attr):
        """If this row is marked for lazy-loading, attempt to load the struct
        that contains the requested property and try again.

        If the requested property requires data from other
        mdtables, all tables will be loaded.
        """
        if self._loaded == LoadState.LazyLoaded:
            if attr in self._class_struct_attrs:
                loader = getattr(self, "_parse" + self._class_struct_attrs[attr])
                loader()
                # If something were to go wrong with loading the correct struct, this
                # would cause a StackOverflow from recursive __getattr__ calls.
                if hasattr(self, attr):
                    return getattr(self, attr)
            elif attr in self._class_struct_attrs_tables:
                # This property requires data from other tables, trigger a full load.
                self._full_loader()
                if hasattr(self, attr):
                    return getattr(self, attr)
        # TODO: more descriptive exception?
        raise AttributeError(attr)

    def parse(self, tables: List["ClrMetaDataTable"], next_row: Optional["MDTableRow"]):
        """
        Parse the row data and set object attributes.  Should only be called after all rows of all tables
        have been initialized, i.e. parse_rows() has been called on each table in the tables list.

            next_row    the next row in the table, used for row lists (e.g. FieldList, MethodList)
        """
        self._parse_struct_asis()
        self._parse_struct_strings()
        self._parse_struct_guids()
        self._parse_struct_blobs()
        self._parse_struct_flags()
        self._parse_struct_enums()
        self._parse_struct_codedindexes(tables, next_row)
        self._parse_struct_indexes(tables, next_row)
        self._parse_struct_lists(tables, next_row)
        self._loaded = LoadState.Loaded

    def _parse_struct_asis(self):
        # if there are any fields to copy as-is
        if hasattr(self.__class__, "_struct_asis"):
            for struct_name, attr_name in self.__class__._struct_asis.items():
                # always define attribute, even if failed to parse
                setattr(self, attr_name, getattr(self.struct, struct_name, None))

    def _parse_struct_strings(self):
        # if strings
        if hasattr(self.__class__, "_struct_strings"):
            for struct_name, attr_name in self.__class__._struct_strings.items():
                # always define attribute, even if failed to parse
                setattr(self, attr_name, None)
                if self._strings is None:
                    logger.warning("failed to fetch string: no strings table")
                    continue

                i = getattr(self.struct, struct_name, None)
                try:
                    s = self._strings.get(i)
                    setattr(self, attr_name, s)
                except UnicodeDecodeError:
                    s = self._strings.get(i, as_bytes=True)
                    logger.warning("string: invalid encoding")
                    setattr(self, attr_name, s)
                except IndexError:
                    logger.warning("failed to fetch string: unable to parse data")

    def _parse_struct_guids(self):
        # if guids
        if hasattr(self.__class__, "_struct_guids"):
            for struct_name, attr_name in self.__class__._struct_guids.items():
                # always define attribute, even if failed to parse
                setattr(self, attr_name, None)
                if self._guids is None:
                    logger.warning("failed to fetch guid: no guid table")
                    continue

                try:
                    g = self._guids.get(getattr(self.struct, struct_name, None))
                    setattr(self, attr_name, g)
                except (IndexError, TypeError):
                    logger.warning("failed to fetch guid: unable to parse data")

    def _parse_struct_blobs(self):
        # if blobs
        if hasattr(self.__class__, "_struct_blobs"):
            for struct_name, attr_name in self.__class__._struct_blobs.items():
                # always define attribute, even if failed to parse
                setattr(self, attr_name, None)
                if self._blobs is None:
                    logger.warning("failed to fetch blob: no blob table")
                    continue
                try:
                    b = self._blobs.get(getattr(self.struct, struct_name, None))
                    setattr(self, attr_name, b)
                except (IndexError, TypeError):
                    logger.warning("failed to fetch blob: unable to parse data")

    def _parse_struct_codedindexes(self, tables, next_row):
        # if coded indexes
        if hasattr(self.__class__, "_struct_codedindexes") and tables:
            for struct_name, (
                attr_name,
                attr_class,
            ) in self.__class__._struct_codedindexes.items():
                # always define attribute, even if failed to parse
                setattr(self, attr_name, None)
                try:
                    o = attr_class(getattr(self.struct, struct_name, None), tables)
                    setattr(self, attr_name, o)
                except (IndexError, TypeError):
                    logger.warning("failed to fetch coded index: unable to parse data")

    def _parse_struct_flags(self):
        # if flags
        if hasattr(self.__class__, "_struct_flags"):
            for struct_name, (attr_name, flag_class) in self.__class__._struct_flags.items():
                # always define attribute, even if failed to parse
                setattr(self, attr_name, None)
                # Set the flags according to the Flags member
                v = getattr(self.struct, struct_name, None)
                if v is None:
                    logger.warning("failed to fetch flag: no data")
                    continue

                try:
                    setattr(self, attr_name, flag_class(v))
                except ValueError:
                    logger.warning("failed to fetch flag: invalid flag data")

    def _parse_struct_enums(self):
        # if enums
        if hasattr(self.__class__, "_struct_enums"):
            for struct_name, (attr_name, enum_class) in self.__class__._struct_enums.items():
                # always define attribute, even if failed to parse
                setattr(self, attr_name, None)
                # Set the value according to the Enum member
                v = getattr(self.struct, struct_name, None)
                if v is None:
                    logger.warning("failed to fetch enum: no data")
                    continue

                try:
                    setattr(self, attr_name, enum_class(v))
                except ValueError:
                    logger.warning("failed to fetch enum: invalid enum data")

    def _parse_struct_indexes(self, tables, next_row):
        # if indexes
        if hasattr(self.__class__, "_struct_indexes") and tables:
            for struct_name, (attr_name, table_name) in self.__class__._struct_indexes.items():
                # always define attribute, even if failed to parse
                setattr(self, attr_name, None)

                table = None
                for t in tables:
                    if t.name == table_name:
                        table = t
                if table:
                    i = getattr(self.struct, struct_name, None)
                    if i is not None and i > 0 and i <= table.num_rows:
                        setattr(self, attr_name, MDTableIndex(table, i))
                    else:
                        logger.warning("failed to fetch index reference: unable to parse data")

    def _parse_struct_lists(self, tables, next_row):
        # if lists
        if hasattr(self.__class__, "_struct_lists") and tables:
            for struct_name, (attr_name, table_name) in self.__class__._struct_lists.items():

                table = None
                for t in tables:
                    if t.name == table_name:
                        table = t

                run: List[MDTableIndex] = []
                # always define attribute, even if failed to parse
                setattr(self, attr_name, run)

                if not table:
                    # target table is not present,
                    # such as is there is no Field table in hello-world.exe,
                    # so the references below must, by defintion, be empty.
                    continue

                run_start_index = getattr(self.struct, struct_name, None)
                if run_start_index is not None:
                    max_row = table.num_rows
                    if next_row is not None:
                        # then we read from the target table,
                        # from the row referenced by this row,
                        # until the row referenced by the next row (`next_row`),
                        # or the end of the table.
                        next_row_reference = getattr(next_row.struct, struct_name, None)
                        run_end_index = max_row
                        if next_row_reference is not None:
                            # row end index is inclusive so row end index must equal next row index minus 1, if less than max row
                            run_end_index = min(next_row_reference - 1, max_row)

                    else:
                        # then we read from the target table,
                        # from the row referenced by this row,
                        # until the end of the table.
                        run_end_index = max_row

                    # when this run starts at the last index,
                    # start == end and end == max_row.
                    # otherwise, if start == end, then run is empty.
                    if run_start_index <= run_end_index:
                        # row indexes are inclusive, so our range goes to end+1
                        for row_index in range(run_start_index, run_end_index + 1):
                            run.append(MDTableIndex(table, row_index))

                setattr(self, attr_name, run)

    def _table_name2num(self, name, tables: List["ClrMetaDataTable"]):
        for t in tables:
            if t.name == name:
                return t.number

    def _clr_coded_index_struct_size(self, tag_bits: int, table_names: Sequence[str]) -> str:
        """
        Given table names and tag bits, checks for the max row count
        size among the given tables and returns "H" if it will fit in
        a word or "I" if not, assuming tag bits number of bits are used
        by tag.

        The returned character can be used in a Structure format or
        passing to struct.pack()
        """
        max_index = 0
        for name in table_names:
            if not name:
                continue

            table_index = enums.MetadataTables[name].value
            table_rowcnt = self._tables_rowcnt[table_index]
            if table_rowcnt is None:
                # the requested table is not present,
                # so it effectively has zero rows.
                table_rowcnt = 0

            max_index = max(max_index, table_rowcnt)

        # if it can fit in a word (minus bits for reference id)
        if max_index <= 2 ** (16 - tag_bits):
            # size is a word
            return "H"
        else:
            # otherwise, size is a dword
            return "I"


# Computing this for each class takes some time, especially if it is done for every row,
# but it *should* remain consistent for any given class and therefore can be cached.
@_functools.lru_cache(None)
def _row_class_struct_attrs(cls: Type[MDTableRow]):
    """Retrieve all possible attributes for a `MDTableRow` class,
    along with their associated struct type.

    Attributes are separated based on whether they can be loaded without data from
    any other tables.
    """
    attrs = {
        attr[0] if isinstance(attr, tuple) else attr: struct
        for struct in MDTableRow.CLASS_ATTRS
        for _, attr in getattr(cls, struct, {}).items()
    }
    attrs_tables = {
        attr[0] if isinstance(attr, tuple) else attr: struct
        for struct in MDTableRow.CLASS_ATTRS_TABLES
        for _, attr in getattr(cls, struct, {}).items()
    }
    return attrs, attrs_tables


class MDTablesStruct(Structure):
    Reserved_1: int
    MajorVersion: int
    MinorVersion: int
    HeapOffsetSizes: int
    Reserved_2: int
    MaskValid: int
    MaskSorted: int


# This type describes the type of row that a table contains.
# For example, the Module table contains ModuleRows,
# so it inherits like this:
#
#     class Module(ClrMetaDataTable[ModuleRow]):
#         ...
#
# This lets us specify that `dnfile.mdtables.Module[0]` is a ModuleRow
# and therefore has properties Name, Generation, etc.
#
# Instances of these types must be subclasses of MDTableRow.
RowType = TypeVar('RowType', bound=MDTableRow)


class MDTableIndex(Generic[RowType]):
    """
    An index into a Metadata Table.

    Attributes:
        table           Table object.
        row_index       Index number of the row.
        row             The referenced row.
    """
    def __init__(self, table: "ClrMetaDataTable[RowType]", row_index: int):
        self.table: Optional["ClrMetaDataTable[RowType]"] = table
        self.row_index: int = row_index

    @property
    def row(self) -> Optional[RowType]:
        if self.table is None:
            return None
        else:
            return self.table.get_with_row_index(self.row_index)


class CodedIndex(MDTableIndex[RowType]):
    """
    Subclasses should be sure to set the following attributes:
      - tag_bits        Number of bits used to specify the table name index.
      - table_names     Candidate list of table names.
    """
    #
    # required properties for subclasses.
    #
    # subclasses must define this property,
    # or __init__ will raise an exception.
    #
    # for example:
    #
    #   class TypeDefOrRef(CodedIndex[...]):
    #       tag_bits = 2
    #       table_names = ("TypeDef", "TypeRef", "TypeSpec")
    #
    tag_bits: int
    table_names: Sequence[str]

    def __init__(self, value, tables: List["ClrMetaDataTable[RowType]"]):
        assert hasattr(self, "tag_bits")
        assert hasattr(self, "table_names")

        table_name = self.table_names[value & (2 ** self.tag_bits - 1)]
        self.row_index = value >> self.tag_bits

        for t in tables:
            if t.name != table_name:
                continue

            self.table = t
            return

        # this may not be a problem, e.g. when ManifestResource Implementation=0
        self.table = None


class ClrMetaDataTable(Generic[RowType]):
    """
    An abstract class for Metadata tables.  Rows can be accessed
     directly like a list with bracket [] syntax.
    Use `get_with_row_index` when you have a Rid/token/row_index,
     since these are 1-indexed.
    Use bracket [] syntax when you want 0-indexing.

    Subclasses should make sure to set the following attributes:
        number
        name
        _row_class
    """
    #
    # required properties for subclasses.
    #
    # subclasses must define this property,
    # or __init__ will raise an exception.
    #
    # for example:
    #
    #   class Module(ClrMetaDataTable[ModuleRow]):
    #       name = "Module"
    #       number = 0
    #       _row_class = ModuleRow
    #
    rva: int
    file_offset: int
    number: int
    name: str
    _row_class: Type[RowType]

    def __init__(
        self,
        tables_rowcounts: List[Optional[int]],
        is_sorted: bool,
        strings_offset_size: int,
        guid_offset_size: int,
        blob_offset_size: int,
        strings_heap: Optional["stream.StringsHeap"],
        guid_heap: Optional["stream.GuidHeap"],
        blob_heap: Optional["stream.BlobHeap"],
        lazy_load=False
    ):
        """
        Given the tables' row counts, sorted flag, and heap info.
        Initialize the following attributes:
            num_rows    The number of rows, according to tables_rowscount.
            row_size    The size, in bytes, of one row.  Calculated from
                        tables_info, tables_rowcounts, and self._format
            is_sorted   Whether the table is sorted, according to tables_info.
            rows        Initialized but not-yet-parsed list of rows.
        """
        assert hasattr(self, "number")
        assert hasattr(self, "name")
        assert hasattr(self, "_row_class")

        self._loaded = LoadState.Unloaded

        # default value
        self.rva: int = 0
        self.file_offset = 0

        num_rows = tables_rowcounts[self.number]
        if num_rows is None:
            # the table doesn't exist, so create the instance, but with zero rows.
            num_rows = 0

        self.is_sorted: bool = is_sorted
        self.num_rows: int = num_rows

        def init_row():
            return self._row_class(
                tables_rowcounts,
                strings_offset_size,
                guid_offset_size,
                blob_offset_size,
                strings_heap,
                guid_heap,
                blob_heap,
            )

        self.rows: List[RowType]
        if lazy_load and num_rows > 0:
            self.rows = _LazyList(self._lazy_parse_rows, num_rows)
            try:
                # `_get_row_size` uses the size of the first row.
                self.rows[0] = init_row()
            except errors.dnFormatError:
                logger.warning("failed to construct %s row %d", self.name, 0)
                # "truncate" the list since the following data is assumed invalid.
                self.rows = []
        else:
            self.rows = []
            for e in range(num_rows):
                try:
                    self.rows.append(init_row())
                except errors.dnFormatError:
                    # this may occur when the offset to a stream is too large.
                    # this probably means invalid data.
                    logger.warning("failed to construct %s row %d", self.name, e)

        # store heap info
        self._strings_heap: Optional["stream.StringsHeap"] = strings_heap
        self._guid_heap: Optional["stream.GuidHeap"] = guid_heap
        self._blob_heap: Optional["stream.BlobHeap"] = blob_heap
        self._strings_offset_size = strings_offset_size
        self._guid_offset_size = guid_offset_size
        self._blob_offset_size = blob_offset_size
        self._tables_rowcounts = tables_rowcounts

        self._table_data: bytes = b""
        self.row_size: int = self._get_row_size()

    def _get_row_size(self):
        if not self.rows:
            return 0
        r = self.rows[0]
        return r.row_size

    def setup_lazy_load(self, table_rva: int, data: bytes, full_loader):
        """Mark this table for lazy-loading.

        `full_loader` will be called if a row property is requested that requires
        loading all mdtables before it can be parsed.
        """
        if not self._loaded == LoadState.Unloaded:
            return
        self.rva = table_rva
        self._table_data = data
        self._full_loader = full_loader
        if len(data) < self.row_size * self.num_rows:
            logger.warning("not enough data to parse %d rows", self.num_rows)
        self._loaded = LoadState.LazyLoaded

    def _lazy_parse_row(self, row, idx):
        """If the row has not been loaded, return a new lazy-loaded row initialized
        with the proper data for the row index.
        """
        if row and row._loaded != LoadState.Unloaded:
            return row

        try:
            row = self._row_class(
                self._tables_rowcounts,
                self._strings_offset_size,
                self._guid_offset_size,
                self._blob_offset_size,
                self._strings_heap,
                self._guid_heap,
                self._blob_heap,
            )
        except errors.dnFormatError:
            # this may occur when the offset to a stream is too large.
            # this probably means invalid data.
            logger.warning("failed to construct %s row %d", self.name, idx)
            # truncate the row list to the current index since the following data is
            # assumed invalid.
            assert isinstance(self.rows, _LazyList)
            self.rows.truncate(idx)
            return None

        row.setup_lazy_load(self._full_loader)
        offset = self.row_size * idx
        if len(self._table_data) < offset + self.row_size:
            logger.warning("not enough data to parse row %d", idx)
            # we could truncate here as well, but regular loading would still be
            # left with a full-length list in the equivalent situation.
            return row
        row.set_data(self._table_data[offset:offset + self.row_size], file_offset=self.file_offset + offset)
        return row

    def _lazy_parse_rows(self, key, row):
        """Convenience function to handle both indexing and slicing of the LazyList."""
        # Guard if called before setup_lazy_load()
        if self._loaded != LoadState.LazyLoaded:
            return row

        if isinstance(key, slice):
            # normally not knowing the length of the list a slice was taken from
            # would be an issue, but we can just keep going until `row` is exhausted.
            return [
                self._lazy_parse_row(row, i)
                for row, i in zip(row, _itertools.count(key.start or 0, key.step or 1))
            ]

        return self._lazy_parse_row(row, key)

    def parse_rows(self, table_rva: int, data: bytes):
        """
        Given a byte sequence containing the rows, add data to each row in the
        self.rows list.  Note that the rows have not been fully parsed until
        parse() is called, which should not happen until all tables have been
        initialized and parse_rows() called on each.
        """
        if self._loaded == LoadState.LazyLoaded:
            # will be handled by lazy evaluation during parse()
            return

        self._table_data = data
        if len(data) < self.row_size * self.num_rows:
            logger.warning("not enough data to parse %d rows", self.num_rows)
            # we can still try to parse some of the rows...

        offset = 0
        # iterate through rows, stopping at num_rows or when there is not enough data left
        for i in range(self.num_rows):
            if len(data) < offset + self.row_size:
                logger.warning("not enough data to parse row %d", i)
                break

            self.rows[i].set_data(
                data[offset:offset + self.row_size], file_offset=self.file_offset + offset
            )
            offset += self.row_size

    def parse(self, tables: List["ClrMetaDataTable"]):
        """
        Fully parse the table, resolving references to heaps, indexes into other
        tables, and coded indexes into other tables.

        NOTE: do not call until ALL tables have been initialized and parse_rows()
        called on each.
        """

        # for each row in table
        for i, row in enumerate(self.rows):
            next_row = None
            if i + 1 < len(self.rows):
                next_row = self.rows[i + 1]

            # fully parse the row
            row.parse(tables, next_row=next_row)
        self._loaded = LoadState.Loaded

    def __getitem__(self, index: int) -> RowType:
        return self.rows[index]

    def __len__(self):
        """
        the actual number of rows parsed,
        as opposed to self.num_rows, which is the declared row count of the table
        """
        return len(self.rows)

    def __iter__(self):
        return iter(self.rows)

    def get_with_row_index(self, row_index: int) -> RowType:
        """
        fetch the row with the given row index.
        remember: row indices, at least those encoded within a .NET file, are 1-based.
        so, you should prefer to use this method when you get a reference to a row.
        use `__getitem__` when you want 0-based indexing.
        """
        return self[row_index - 1]


class ClrResource(abc.ABC):
    def __init__(self, name: str, public: bool = False, private: bool = False):
        self.name: str = name
        self.public: bool = public
        self.private: bool = private
        self.data: Optional[bytes] = None

    def set_data(self, data: bytes):
        self.data = data

    @abc.abstractmethod
    def parse(self):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
dnfile, .NET Executable file parser


REFERENCES

    https://www.ntcore.com/files/dotnetformat.htm
    https://referencesource.microsoft.com/System.AddIn/System/Addin/MiniReflection/MetadataReader/Metadata.cs.html
    https://asmresolver.readthedocs.io/en/latest/peimage/dotnet.html#metadata-streams


Copyright (c) 2020-2022 MalwareFrank
"""

__author__ = """MalwareFrank"""
__version__ = "0.15.1"

import copy as _copymod
import codecs
import struct as _struct
import logging
from typing import Dict, List, Optional

from pefile import PE as _PE
from pefile import DIRECTORY_ENTRY, MAX_SYMBOL_EXPORT_COUNT, Dump, Structure, DataContainer, PEFormatError

from . import base, enums, errors, stream, resource

logger = logging.getLogger(__name__)
CLR_METADATA_SIGNATURE = 0x424A5342


# These come from the great article[1] which contains great insights on
# working with unicode in both Python 2 and 3.
# [1]: http://python3porting.com/problems.html
def handler(err):
    start = err.start
    end = err.end
    values = [
        ("\\u{0:04x}" if ord(err.object[i]) > 255 else "\\x{0:02x}", ord(err.object[i]))
        for i in range(start, end)
    ]
    return (u"".join([elm[0].format(elm[1]) for elm in values]), end)


codecs.register_error("backslashreplace_", handler)


class dnPE(_PE):
    def add_warning(self, msg):
        self._warnings.append(msg)

    def __init__(
        self,
        name=None,
        data=None,
        fast_load=None,
        max_symbol_exports=MAX_SYMBOL_EXPORT_COUNT,
        clr_lazy_load=False,
    ):
        self._warnings = list()
        self.clr_lazy_load = clr_lazy_load
        super().__init__(name, data, fast_load)

    def dump_info(self, dump=None, encoding="utf-8"):
        """
        Dump all the PE and CLR header information into human readable string.
        """
        if dump is None:
            dump = Dump()

        super().dump_info(dump, encoding)

        if not hasattr(self, "net") or not self.net:
            return dump.get_text()

        #### CLR
        # directory entry
        dump.add_header("CLR (.NET)")
        dump.add_lines(self.net.struct.dump())
        dump.add_newline()

        # metadata
        if hasattr(self.net, "metadata") and self.net.metadata:
            dump.add_lines(self.net.metadata.struct.dump(), indent=2)
            dump.add_newline()
            # Streams
            if (
                hasattr(self.net.metadata, "streams_list")
                and self.net.metadata.streams_list
            ):
                for stream_ in self.net.metadata.streams_list:
                    dump.add_lines(stream_.struct.dump(), indent=4)
                    dump.add_newline()

        # Metadata Tables
        if hasattr(self.net, "mdtables") and self.net.mdtables:
            dump.add_header("CLR (.NET) Metadata Tables")
            for s in self.net.metadata.streams_list:
                if isinstance(s, stream.MetaDataTables):
                    s: stream.MetaDataTables
                    dump.add_lines(s.struct.dump())
                    dump.add_newline()
                    if hasattr(s, "tables_list") and s.tables_list:
                        for t in s.tables_list:
                            for label, value in (
                                ("RVA", hex(t.rva)),
                                ("TableName", t.name),
                                ("TableNumber", t.number),
                                ("IsSorted", t.is_sorted),
                                ("NumRows", t.num_rows),
                                ("RowSize", t.row_size),
                            ):
                                dump.add_line(
                                    "{0:<20}{1}".format(label  + ":", str(value)),
                                    indent=2,
                                )
                            dump.add_newline()

        return dump.get_text()

    def get_warnings(self):
        """
        Returns a copy of the list of warning messages.
        """
        result = _copymod.deepcopy(super().get_warnings())
        result.extend(self._warnings)
        return result

    def __parse__(self, fname, data, fast_load):
        super().__parse__(fname, data, fast_load)

        # NOTE: .NET loaders ignores NumberOfRvaAndSizes
        #   We check this elsewhere, but note it here.
        #   example: 1d41308bf4148b4c138f9307abc696a6e4c05a5a89ddeb8926317685abb1c241

    def parse_data_directories(
        self, directories=None, forwarded_exports_only=False, import_dllnames_only=False
    ):
        super().parse_data_directories(
            directories, forwarded_exports_only, import_dllnames_only
        )

        directory_parsing = (
            ("IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR", self.parse_clr_structure),
        )

        # Indicate no .NET object parsed
        self.net = None

        if self.__data__ is None:
            logger.warning("not a .NET module: failed to read data")
            return

        opt_header = getattr(self, "OPTIONAL_HEADER", None)
        if opt_header is None:
            logger.warning("not a .NET module: no optional header")
            return

        if directories is not None:
            if not isinstance(directories, (tuple, list)):
                directories = [directories]

        for entry in directory_parsing:
            try:
                directory_index = DIRECTORY_ENTRY[entry[0]]
                dir_entry = opt_header.DATA_DIRECTORY[directory_index]
            except IndexError:
                break

            # Only process all the directories if no individual ones have
            # been chosen
            #
            if directories is None or directory_index in directories:

                if dir_entry.VirtualAddress:
                    value = entry[1](dir_entry.VirtualAddress, dir_entry.Size)
                    if value:
                        setattr(self, entry[0][6:], value)
                        if entry[0] == "IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR":
                            # create shortcut for .NET/CLR data
                            self.net = value

            if (
                (directories is not None)
                and isinstance(directories, list)
                and (entry[0] in directories)
            ):
                directories.remove(directory_index)

        # NOTE: .NET loaders ignores NumberOfRvaAndSizes, so attempt to parse anyways
        #   example: 1d41308bf4148b4c138f9307abc696a6e4c05a5a89ddeb8926317685abb1c241
        attr_name = "DIRECTORY_ENTRY_COM_DESCRIPTOR"
        if not hasattr(self, attr_name):
            dir_entry_size = Structure(self.__IMAGE_DATA_DIRECTORY_format__).sizeof()
            dd_offset = (
                opt_header.get_file_offset() + opt_header.sizeof()
            )
            clr_entry_offset = dd_offset + (
                DIRECTORY_ENTRY["IMAGE_DIRECTORY_ENTRY_COM_DESCRIPTOR"] * dir_entry_size
            )
            data = self.__data__[clr_entry_offset:clr_entry_offset + dir_entry_size]
            dir_entry = self.__unpack_data__(
                self.__IMAGE_DATA_DIRECTORY_format__, data, file_offset=clr_entry_offset
            )

            # if COM entry appears valid
            if dir_entry is not None and dir_entry.VirtualAddress:
                # try to parse the .NET CLR directory
                value = self.parse_clr_structure(
                    dir_entry.VirtualAddress, dir_entry.Size
                )
                if value:
                    setattr(self, attr_name, value)
                    # create shortcut for .NET/CLR data
                    self.net = value

    def parse_clr_structure(self, rva, size) -> Optional["ClrData"]:
        try:
            return ClrData(self, rva, size, self.clr_lazy_load)
        except errors.dnFormatError as e:
            logger.warning("failed to parse CLR data: %s",  e)
            return None


class ClrMetaDataStruct(Structure):
    Signature: int
    MajorVersion: int
    MinorVersion: int
    Reserved: int
    VersionLength: int
    Version: int
    Flags: int
    NumberOfStreams: int


class ClrMetaData(DataContainer):
    """Holds CLR (.NET) MetaData.

    struct:         IMAGE_CLR_METADATA structure
    streams:        Dictionary to access streams by name (bytes)
    streams_list:   List of streams in order of entry in header
    """

    rva: int
    struct: ClrMetaDataStruct
    streams: Dict[bytes, base.ClrStream]
    streams_list: List[base.ClrStream]

    _format = (
        "IMAGE_CLR_METADATA",
        [
            "I,Signature",
            "H,MajorVersion",
            "H,MinorVersion",
            "I,Reserved",
            "I,VersionLength",
            # '?,Version',
            # 'H,Flags',
            # 'H,NumberOfStreams',
        ],
    )
    #### MetaData section
    #
    # dd    Signature
    # dw    MajorVersion
    # dw    MinorVersion
    # dd    Reserved
    # dd    Length
    # var   Version
    # dw    Flags
    # dw    NumberOfStreams
    # var   StreamHeaders

    def __init__(self, pe: dnPE, rva: int, size: int, lazy_load=False):
        """
        Given a dnPE object, MetaData RVA and MetaData Size.
        Raises dnFormatError if encounter problems parsing.
        """
        metadata_rva = rva

        # The metadata RVA, used for stream offsets
        self.rva = rva

        # dynamically create metadata header structure
        struct_format = _copymod.deepcopy(self.__class__._format)
        struct_data = pe.get_data(rva, size)
        if len(struct_data) < size:
            raise errors.dnFormatError(
                "Invalid CLR MetaData Structure size. Can't read %d "
                "bytes at RVA: 0x%x" % (size, rva)
            )
        # check signature
        sig = _struct.unpack_from("<I", struct_data)[0]
        if sig != CLR_METADATA_SIGNATURE:
            raise errors.dnFormatError(
                "Invalid CLR MetaData Signature at 0x%x. Expected 0x%x but "
                "got 0x%x" % (rva, CLR_METADATA_SIGNATURE, sig)
            )
        # parse struct so that we can get the version length
        metadata_struct = ClrMetaDataStruct(
            format=struct_format,
            file_offset=pe.get_offset_from_rva(metadata_rva)
        )
        metadata_struct.__unpack__(struct_data)
        # metadata_struct = pe.__unpack_data__(
        #     struct_format, struct_data, pe.get_offset_from_rva(metadata_rva)
        # )
        # add variable-length version field
        if metadata_struct.VersionLength > 0:
            struct_format[1].append(
                "{0}s,Version".format(metadata_struct.VersionLength)
            )
        # add Flags
        struct_format[1].append("H,Flags")
        # add NumberOfStreams
        struct_format[1].append("H,NumberOfStreams")

        # re-parse metadata header structure
        metadata_struct = ClrMetaDataStruct(
            format=struct_format,
            file_offset=pe.get_offset_from_rva(metadata_rva)
        )
        struct_size = metadata_struct.sizeof()
        struct_data = pe.get_data(metadata_rva, struct_size)
        if len(struct_data) < struct_size:
            raise errors.dnFormatError(
                "unable to read full CLR metadata structure, expected {} got {}".format(
                    struct_size, len(struct_data)
                )
            )
        metadata_struct.__unpack__(struct_data)

        self.struct = metadata_struct

        if metadata_struct.NumberOfStreams > 0:
            # parse the streams table
            streams_table_rva = metadata_rva + struct_size
            self.parse_stream_table(pe, streams_table_rva)

            # parse each stream
            for s in self.streams_list:
                try:
                    s.parse(self.streams_list, lazy_load=lazy_load)
                except (errors.dnFormatError, PEFormatError) as e:
                    # other streams may parse, so add to warnings and continue
                    pe.add_warning("Unable to parse stream {!r}".format(s.struct.Name))
                    pe.add_warning(str(e))
                    logger.warning("unable to parse stream: %s: %s", s.struct.Name, e)

    def parse_stream_table(self, pe: dnPE, streams_table_rva):
        streams_list = list()
        streams_dict = dict()
        # pointer to current stream's table entry
        stream_entry_rva = streams_table_rva
        for i in range(self.struct.NumberOfStreams):
            stream = ClrStreamFactory.createStream(pe, stream_entry_rva, self.rva)
            if not stream:
                logger.warning("Invalid .NET stream: {}".format(i + 1))
                pe.add_warning("Invalid .NET stream: {}".format(i + 1))
                # assume this throws off further parsing, so stop
                break

            streams_list.append(stream)
            name = stream.struct.Name
            if name in streams_dict:
                # if a stream with this name already exists.
                # this is not fatal, just unusual.
                pe.add_warning("Duplicate .NET stream name '{!r}'".format(name))
                logger.warning("Duplicate .NET stream name: %s", name)

            # dotnet uses the last encountered stream with a given name,
            # see: https://github.com/malwarefrank/dnfile/issues/19#issuecomment-992754448
            # and test_invalid_streams.py::test_duplicate_stream
            streams_dict[name] = stream
            # move to next entry in streams table
            stream_entry_rva += stream.stream_table_entry_size()

        self.streams = streams_dict
        self.streams_list = streams_list


class ClrStruct(Structure):
    cb: int
    MajorRuntimeVersion: int
    MinorRuntimeVersion: int
    MetaDataRva: int
    MetaDataSize: int
    Flags: int
    EntryPointTokenOrRva: int
    ResourcesRva: int
    ResourcesSize: int
    StrongNameSignatureRva: int
    StrongNameSignatureSize: int
    CodeManagerTableRva: int
    CodeManagerTableSize: int
    VTableFixupsRva: int
    VTableFixupsSize: int
    ExportAddressTableJumpsRva: int
    ExportAddressTableJumpsSize: int
    ManagedNativeHeaderRva: int
    ManagedNativeHeaderSize: int


class ClrData(DataContainer):
    """Holds CLR (.NET) header data.

    struct:         IMAGE_NET_DIRECTORY structure
    metadata:       ClrMetaData or None
    strings:        stream.StringsHeap or None
    user_strings:   stream.UserStringsHeap or None
    guids:          stream.GuidHeap or None
    blobs:          stream.BlobHeap or None
    mdtables:       stream.MetaDataTables or None
    Flags:          enums.ClrHeaderFlags or None
    """

    struct: ClrStruct
    metadata: Optional[ClrMetaData]
    strings: Optional[stream.StringsHeap]
    user_strings: Optional[stream.UserStringHeap]
    guids: Optional[stream.GuidHeap]
    blobs: Optional[stream.BlobHeap]
    mdtables: Optional[stream.MetaDataTables]
    Flags: Optional[enums.ClrHeaderFlags]

    _resources: Optional[List[base.ClrResource]]

    @property
    def resources(self) -> List[base.ClrResource]:
        if self._resources is None:
            self._init_resources(getattr(self, "_pe"))
            assert self._resources is not None
        return self._resources

    # Structure description from:
    # http://www.ntcore.com/files/dotnetformat.htm
    _format = (
        "IMAGE_NET_DIRECTORY",
        (
            "I,cb",
            "H,MajorRuntimeVersion",
            "H,MinorRuntimeVersion",
            "I,MetaDataRva",
            "I,MetaDataSize",
            "I,Flags",
            "I,EntryPointTokenOrRva",
            "I,ResourcesRva",
            "I,ResourcesSize",
            "I,StrongNameSignatureRva",
            "I,StrongNameSignatureSize",
            "I,CodeManagerTableRva",
            "I,CodeManagerTableSize",
            "I,VTableFixupsRva",
            "I,VTableFixupsSize",
            "I,ExportAddressTableJumpsRva",
            "I,ExportAddressTableJumpsSize",
            "I,ManagedNativeHeaderRva",
            "I,ManagedNativeHeaderSize",
        ),
    )

    def __init__(self, pe: dnPE, rva: int, size: int, lazy_load: bool):
        """
        Given dnPE object, .NET header RVA and header size.
        Raises dnFormatError if problems parsing.
        """
        data = pe.get_data(rva, size)

        try:
            clr_struct = ClrStruct(
                self._format, file_offset=pe.get_offset_from_rva(rva)
            )
            clr_struct.__unpack__(data)
        except PEFormatError:
            # raise exception because we can't do anything halfway here.
            raise errors.dnFormatError(
                "Invalid CLR Structure information. Can't read "
                "data at RVA: 0x%x" % rva
            )

        # set structure member
        self.struct = clr_struct
        # parse metadata
        metadata_rva = clr_struct.MetaDataRva
        metadata_size = clr_struct.MetaDataSize

        # shortcuts are always defined, even if None, for easier code completion in IDEs
        self.metadata = None
        self.strings = None
        self.user_strings = None
        self.guids = None
        self.blobs = None
        self.mdtables = None
        self._resources = None
        self.Flags = None

        try:
            self.metadata = ClrMetaData(pe, metadata_rva, metadata_size, lazy_load)
        except (errors.dnFormatError, PEFormatError) as e:
            logger.warning("failed to parse .NET metadata: %s", e)
            return

        # create shortcuts for streams
        # dotnet runtime uses the last instance of a type,
        # see: https://github.com/malwarefrank/dnfile/issues/19#issuecomment-992754448
        # and test: test_invalid_streams.py::test_duplicate_stream
        for s in self.metadata.streams_list:
            if isinstance(s, stream.StringsHeap):
                self.strings = s
            elif isinstance(s, stream.UserStringHeap):
                self.user_strings = s
            elif isinstance(s, stream.GuidHeap):
                self.guids = s
            elif isinstance(s, stream.BlobHeap):
                self.blobs = s
            elif isinstance(s, stream.MetaDataTables):
                self.mdtables: stream.MetaDataTables = s

        # Set the flags according to the Flags member
        flags_object = enums.ClrHeaderFlags(clr_struct.Flags)
        self.Flags = flags_object

        if not lazy_load:
            self._init_resources(pe)
        else:
            # store the dnPE reference for lazy-loading
            setattr(self, "_pe", pe)

    def _init_resources(self, pe):
        """Parse and initialize assembly resources.

        This is separate from `ClrData.__init__` to allow for mdtable lazy-loading since
        parsing `ManifestResourceRow.Implementation` requires all tables to be loaded.
        """
        self._resources = []
        # parse the resources
        if self.struct.ResourcesRva > 0 and self.mdtables and self.mdtables.ManifestResource and self.mdtables.ManifestResource.num_rows > 0:
            # for each row
            for row in self.mdtables.ManifestResource.rows:
                # TODO: handle external resources
                if row.Implementation is None:
                    # internal resource, embedded in this file
                    rva = self.struct.ResourcesRva + row.Offset
                    try:
                        buf = pe.get_data(rva, 4)
                    except PEFormatError as e:
                        # warn
                        pe.add_warning("CLR resource parse error, expected more data at rva 0x{:02x}".format(rva))
                        continue
                    if not buf or len(buf) < 4:
                        # warn
                        pe.add_warning("CLR resource parse error, expected at least 4 bytes at rva 0x{:02x}".format(rva))
                        continue
                    size = int.from_bytes(buf, byteorder="little")
                    rsrc_rva = rva + 4
                    try:
                        rdata = pe.get_data(rsrc_rva, size)
                    except PEFormatError as e:
                        # warn
                        pe.add_warning("CLR resource parse error, expected more data at rva 0x{:02x}".format(rsrc_rva))
                        continue
                    if not rdata or len(rdata) < size:
                        pe.add_warning("CLR resource parse error, expected more data at rva 0x{:02x}".format(rsrc_rva))
                        continue
                    res = resource.InternalResource(row.Name, row.Flags.mrPublic, row.Flags.mrPrivate)
                    res.rva = rsrc_rva
                    res.size = size
                    res.data = rdata
                    self._resources.append(res)
            for rsrc in self._resources:
                try:
                    rsrc.parse()
                except errors.dnFormatError as e:
                    if isinstance(rsrc, resource.InternalResource):
                        pe.add_warning("CLR resource parse error for '{}' at 0x{:02x}: {}".format(rsrc.name, rsrc.rva, str(e)))
                    else:
                        pe.add_warning("CLR resource parse error for '{}': {}".format(rsrc.name, str(e)))


class ClrStreamFactory(object):
    _name_type_map = {
        b"#~": stream.MetaDataTables,
        b"#-": stream.MetaDataTables,
        b"#Schema": stream.MetaDataTables,
        b"#Strings": stream.StringsHeap,
        b"#GUID": stream.GuidHeap,
        b"#Blob": stream.BlobHeap,
        b"#US": stream.UserStringHeap,
    }
    _template_format = (
        "IMAGE_CLR_STREAM",
        [
            "I,Offset",
            "I,Size",
            # '?,Name',
        ],
    )

    @classmethod
    def createStream(
        cls, pe: dnPE, stream_entry_rva: int, metadata_rva: int
    ) -> Optional[base.ClrStream]:
        # start with structure template
        struct_format = _copymod.deepcopy(cls._template_format)
        # read name
        name = pe.get_string_at_rva(stream_entry_rva + 8)
        if name is None:
            logger.warning("failed to read stream name")
            return None

        # round field length up to next 4-byte boundary.  Remember the NULL byte at end.
        name_len = len(name) + (4 - (len(name) % 4))
        # add name field to structure
        struct_format[1].append("{0}s,Name".format(name_len))
        # parse structure
        stream_struct = base.StreamStruct(
            struct_format,
            file_offset=pe.get_offset_from_rva(stream_entry_rva)
        )
        struct_size = stream_struct.sizeof()
        struct_data = pe.get_data(stream_entry_rva, struct_size)
        stream_struct.__unpack__(struct_data)
        # remove trailing NULLs from name
        stream_struct.Name = stream_struct.Name.rstrip(b"\x00")
        stream_rva = metadata_rva + stream_struct.Offset
        stream_data = pe.get_data(
            stream_rva, stream_struct.Size
        )
        name = stream_struct.Name
        # use GenericStream for any non-standard streams
        stream_class = cls._name_type_map.get(name, stream.GenericStream)
        try:
            # construct stream, like stream.StreagsHeap ctor or GenericStream ctor
            s = stream_class(metadata_rva, stream_struct, stream_data)
            s.file_offset = pe.get_offset_from_rva(stream_rva)
        except errors.dnFormatError as e:
            logger.warning("failed to parse stream: %s", e)
            return None
        else:
            return s



import pytest
import fixtures

import dnfile
from dnfile.mdtable import TypeRefRow, AssemblyRefRow


def test_metadata():
    path = fixtures.get_data_path_by_name("hello-world.exe")

    dn = dnfile.dnPE(path)
    assert dn.net is not None
    assert dn.net.metadata is not None

    dn.net.metadata.struct.Signature == 0x424A5342
    dn.net.metadata.struct.MajorVersion == 1
    dn.net.metadata.struct.MinorVersion == 1
    dn.net.metadata.struct.Version == "v4.0.30319"
    dn.net.metadata.struct.Flags == 0x0
    dn.net.metadata.struct.NumberOfStreams == 5


def test_streams():
    path = fixtures.get_data_path_by_name("hello-world.exe")

    dn = dnfile.dnPE(path)
    assert dn.net is not None
    assert dn.net.metadata is not None

    assert hasattr(dn.net.metadata, "streams")
    assert b"#~" in dn.net.metadata.streams
    assert dn.net.metadata.streams[b"#~"].get_file_offset() == 0x2d4

    # strings used by #~
    assert b"#Strings" in dn.net.metadata.streams
    assert hasattr(dn.net, "strings")
    assert dn.net.strings.get_file_offset() == 0x3d8

    # "user strings"
    assert b"#US" in dn.net.metadata.streams
    assert hasattr(dn.net, "user_strings")
    assert dn.net.user_strings.get_file_offset() == 0x4dc

    assert b"#GUID" in dn.net.metadata.streams
    assert hasattr(dn.net, "guids")
    assert dn.net.guids.get_file_offset() == 0x4f8

    assert b"#Blob" in dn.net.metadata.streams
    assert hasattr(dn.net, "blobs")
    assert dn.net.blobs.get_file_offset() == 0x508

    assert b"#Foo" not in dn.net.metadata.streams
    assert not hasattr(dn.net, "foo")


def test_guids():
    path = fixtures.get_data_path_by_name("hello-world.exe")

    dn = dnfile.dnPE(path)
    assert dn.net is not None
    assert b"#GUID" in dn.net.metadata.streams
    assert hasattr(dn.net, "guids")

    assert dn.net.guids.get(0) is None
    assert dn.net.guids.get(1).value == b"\x8c\x8b\xc5\x48\xff\x24\x91\x45\x9e\xc8\x94\xbf\xea\xbd\x9f\x3e"


def test_tables():
    path = fixtures.get_data_path_by_name("hello-world.exe")

    dn = dnfile.dnPE(path)
    assert dn.net is not None

    for table in ["Module", "TypeRef", "TypeDef", "MethodDef", "Param", "MemberRef", "CustomAttribute", "Assembly", "AssemblyRef"]:
        assert hasattr(dn.net.mdtables, table)

    assert not hasattr(dn.net.mdtables, "foo")

    tables = dict()
    tables["Module"] = {
        "RVA": 0x2110,
        "TableName": "Module",
        "TableNumber": 0,
        "IsSorted": False,
        "NumRows": 1,
        "RowSize": 10,
        "file_offset": 0x310,
    }
    tables["TypeRef"] = {
        "RVA": 0x211a,
        "TableName": "TypeRef",
        "TableNumber": 1,
        "IsSorted": False,
        "NumRows": 6,
        "RowSize": 6,
        "file_offset": 0x31a,
    }
    tables["TypeDef"] = {
        "RVA": 0x213e,
        "TableName": "TypeDef",
        "TableNumber": 2,
        "IsSorted": False,
        "NumRows": 2,
        "RowSize": 14,
        "file_offset": 0x33e,
    }
    tables["MethodDef"] = {
        "RVA": 0x215a,
        "TableName": "MethodDef",
        "TableNumber": 6,
        "IsSorted": False,
        "NumRows": 2,
        "RowSize": 14,
        "file_offset": 0x35a,
    }
    tables["Param"] = {
        "RVA": 0x2176,
        "TableName": "Param",
        "TableNumber": 8,
        "IsSorted": False,
        "NumRows": 1,
        "RowSize": 6,
        "file_offset": 0x376,
    }
    tables["MemberRef"] = {
        "RVA": 0x217c,
        "TableName": "MemberRef",
        "TableNumber": 10,
        "IsSorted": False,
        "NumRows": 5,
        "RowSize": 6,
        "file_offset": 0x37c,
    }
    tables["CustomAttribute"] = {
        "RVA": 0x219a,
        "TableName": "CustomAttribute",
        "TableNumber": 12,
        "IsSorted": True,
        "NumRows": 3,
        "RowSize": 6,
        "file_offset": 0x39a,
    }
    tables["Assembly"] = {
        "RVA": 0x21ac,
        "TableName": "Assembly",
        "TableNumber": 32,
        "IsSorted": False,
        "NumRows": 1,
        "RowSize": 22,
        "file_offset": 0x3ac,
    }
    tables["AssemblyRef"] = {
        "RVA": 0x21c2,
        "TableName": "AssemblyRef",
        "TableNumber": 35,
        "IsSorted": False,
        "NumRows": 1,
        "RowSize": 20,
        "file_offset": 0x3c2,
    }

    for table in dn.net.mdtables.tables_list:
        table: dnfile.base.ClrMetaDataTable
        ref_table = tables[table.name]
        assert table.name in tables
        assert table.rva == ref_table.get("RVA", None)
        assert table.name == ref_table.get("TableName", None)
        assert table.number == ref_table.get("TableNumber", None)
        assert table.is_sorted == ref_table.get("IsSorted", None)
        assert table.num_rows == ref_table.get("NumRows", None)
        assert table.row_size == ref_table.get("RowSize", None)
        assert table.file_offset == ref_table.get("file_offset")


def test_module():
    path = fixtures.get_data_path_by_name("hello-world.exe")

    dn = dnfile.dnPE(path)
    assert dn.net is not None

    assert dn.net.mdtables.Module[0].Name == "1-hello-world.exe"


def test_typedef_extends():
    path = fixtures.get_data_path_by_name("hello-world.exe")

    dn = dnfile.dnPE(path)
    assert dn.net is not None

    typedefs = dn.net.mdtables.TypeDef
    assert typedefs[0].TypeName == "<Module>"
    assert typedefs[1].TypeName == "HelloWorld"

    #   .class public auto ansi beforefieldinit HelloWorld
    #      extends [mscorlib]System.Object

    extends = typedefs[1].Extends
    assert extends.table is not None
    assert extends.table.name == "TypeRef"
    assert extends.row_index == 5

    superclass = extends.row
    assert isinstance(superclass, TypeRefRow)
    assert superclass.TypeNamespace == "System"
    assert superclass.TypeName == "Object"

    assert superclass.ResolutionScope.table is not None
    assert superclass.ResolutionScope.table.name == "AssemblyRef"
    assembly = superclass.ResolutionScope.row
    assert isinstance(assembly, AssemblyRefRow)
    assert assembly.Name == "mscorlib"


def test_typedef_members():
    path = fixtures.get_data_path_by_name("ModuleCode_x86.exe")

    dn = dnfile.dnPE(path)
    assert dn.net is not None

    typedefs = dn.net.mdtables.TypeDef
    assert typedefs[0].TypeName == "<Module>"
    assert typedefs[5].TypeName == "ModuleLoadException"

    assert len(typedefs[0].FieldList) == 52
    assert len(typedefs[5].FieldList) == 1

    assert len(typedefs[0].MethodList) == 76
    assert len(typedefs[5].MethodList) == 3

    assert typedefs[0].MethodList[0].row.Name == "rc4_init"
    assert typedefs[5].MethodList[0].row.Name == ".ctor"


def test_typedef_methodlist():
    path = fixtures.get_data_path_by_name("EmptyClass_x86.exe")

    dn = dnfile.dnPE(path)
    assert dn.net is not None

    typedefs = dn.net.mdtables.TypeDef
    assert typedefs[0].TypeName == "<Module>"
    assert typedefs[0].MethodList is not None
    assert len(typedefs[0].MethodList) == 1
    assert typedefs[0].MethodList[0].row.Name == "_mainCRTStartup"


def test_method_params():
    path = fixtures.get_data_path_by_name("hello-world.exe")

    dn = dnfile.dnPE(path)
    assert dn.net is not None

    methods = dn.net.mdtables.MethodDef
    assert methods[0].Name == "Main"
    assert methods[1].Name == ".ctor"

    # default void Main (string[] args)  cil managed
    assert len(methods[0].ParamList) == 1
    # instance default void '.ctor' ()  cil managed
    assert len(methods[1].ParamList) == 0

    assert methods[0].ParamList[0].row is not None
    assert methods[0].ParamList[0].row.Name == "args"


def test_ignore_NumberOfRvaAndSizes():
    # .NET loaders ignores NumberOfRvaAndSizes, so attempt to parse anyways
    path = fixtures.DATA / "1d41308bf4148b4c138f9307abc696a6e4c05a5a89ddeb8926317685abb1c241"
    if not path.exists():
        raise pytest.xfail("test file 1d41308bf41... (DANGER: malware) not found in test fixtures")

    dn = dnfile.dnPE(path)
    assert hasattr(dn, "net") and dn.net is not None
    assert hasattr(dn.net, "metadata") and dn.net.metadata is not None


def test_flags():
    path = fixtures.get_data_path_by_name("hello-world.exe")

    dn = dnfile.dnPE(path)
    assert dn.net is not None

    # class HelloWorld
    cls = dn.net.mdtables.TypeDef.get_with_row_index(2)

    # these are enums from CorTypeSemantics
    assert cls.Flags.tdClass is True
    assert cls.Flags.tdInterface is False

    # these are flags from CorTypeAttrFlags
    assert cls.Flags.tdBeforeFieldInit is True
    assert cls.Flags.tdAbstract is False


def test_heap_items():

    # HeapItem
    buf = b"1234567890"
    item = dnfile.base.HeapItem(b"1234567890")
    assert item.raw_data == b"1234567890"
    assert item.value_bytes() == buf
    assert item == b"1234567890"
    assert item.raw_size == len(b"1234567890")
    assert item.rva is None
    assert item.value is None

    item = dnfile.base.HeapItem(b"1234567890", rva=0x400)
    assert item.rva == 0x400

    # HeapItemString
    buf = b"1234567890"
    buf_decoded = buf.decode("utf-8")
    str_item = dnfile.stream.HeapItemString(buf)
    assert str_item.raw_data == buf
    assert str_item.encoding == "utf-8"
    assert str_item.value == buf_decoded
    assert str_item.value_bytes() == buf
    assert str_item == buf_decoded

    # HeapItemBinary
    buf_with_compressed_int_len = b"\x0a1234567890"
    buf = b"1234567890"
    bin_item = dnfile.stream.HeapItemBinary(buf_with_compressed_int_len)
    assert bin_item.raw_data == buf_with_compressed_int_len
    assert bin_item.raw_size == len(buf_with_compressed_int_len)
    assert bin_item.item_size == len(buf)
    assert bin_item.value_bytes() == buf
    assert bin_item == buf

    # UserString
    buf_with_flag = b"\x151\x002\x003\x004\x005\x006\x007\x008\x009\x000\x00\x00"
    us_str = "1234567890"
    us_item = dnfile.stream.UserString(buf_with_flag)
    assert us_item.raw_data == buf_with_flag
    assert us_item.value_bytes() == buf_with_flag[1:-1]
    assert us_item.value == us_str
    assert us_item == us_str

    # GUID
    guid_bytes = b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"
    guid_item = dnfile.stream.HeapItemGuid(guid_bytes)
    assert guid_item is not None
    assert guid_item.raw_data == guid_bytes
    assert guid_item.value_bytes() == guid_bytes
    assert guid_item.value == guid_bytes
    assert str(guid_item) == "03020100-0504-0706-0809-0a0b0c0d0e0f"



import fixtures

import dnfile
from dnfile.utils import LazyList
from dnfile.mdtable import TypeRefRow, MemberRefRow


def test_lazy_loading():
    path = fixtures.get_data_path_by_name("hello-world.exe")

    dn = dnfile.dnPE(path, clr_lazy_load=True)
    dn.parse_data_directories()

    assert dn.net
    assert dn.net.mdtables

    assert dn.net.mdtables.MemberRef
    assert isinstance(dn.net.mdtables.MemberRef.rows, LazyList)

    # LazyList is initialized as a list of `None`.
    # list.__getitem__ bypasses lazy loading.
    assert list.__getitem__(dn.net.mdtables.MemberRef.rows, 1) is None
    memref_row = dn.net.mdtables.MemberRef.rows[1]
    assert isinstance(memref_row, MemberRefRow)

    assert memref_row.Name
    # __dict__ access bypasses lazy loading.
    assert "Signature" not in memref_row.__dict__
    assert memref_row.Signature
    assert "Signature" in memref_row.__dict__

    # The first item of the rows list is a special case, because it is
    # initialized without data before lazy-loading is setup.
    # Make sure that it behaves the same as any other row when accessed.
    # We can also verify that slicing works, since it too needs to be
    # handled appropriately.
    typeref_row = dn.net.mdtables.TypeRef.rows[:3:2][0]
    assert isinstance(typeref_row, TypeRefRow)

    assert "ResolutionScope" not in typeref_row.__dict__
    # TypeRefRow.Class should trigger a full load of all tables and rows.
    assert memref_row.Class
    # This should not have been lazy-loaded, but will still have been loaded
    # because of accessing the Class attribute.
    assert "ResolutionScope" in typeref_row.__dict__

    # _resources is the underlying lazy-loaded field for the ClrResource list.
    assert dn.net._resources is None
    assert dn.net.resources is not None
    assert dn.net._resources is not None


def test_non_lazy_loading():
    path = fixtures.get_data_path_by_name("hello-world.exe")

    dn = dnfile.dnPE(path, clr_lazy_load=False)
    dn.parse_data_directories()

    assert dn.net
    assert dn.net.mdtables

    assert dn.net.mdtables.MemberRef
    assert not isinstance(dn.net.mdtables.MemberRef.rows, LazyList)

    # list.__getitem__ would bypass any lazy loading.
    memref_row = list.__getitem__(dn.net.mdtables.MemberRef.rows, 1)
    assert isinstance(memref_row, MemberRefRow)
    # __dict__ access would bypass any lazy loading.
    assert "Signature" in memref_row.__dict__

    typeref_row = dn.net.mdtables.TypeRef.rows[0]
    assert isinstance(typeref_row, TypeRefRow)
    assert "ResolutionScope" in typeref_row.__dict__

    # _resources is the underlying field that would be lazy-loaded.
    assert dn.net._resources is not None



# -*- coding: utf-8 -*-
"""
.NET Metadata Tables Coded Indexes


REFERENCES

    https://www.ntcore.com/files/dotnetformat.htm
    https://referencesource.microsoft.com/System.AddIn/System/Addin/MiniReflection/MetadataReader/Metadata.cs.html#123


Copyright (c) 2020-2022 MalwareFrank
"""

from typing import TYPE_CHECKING, Union

from .base import CodedIndex

if TYPE_CHECKING:
    from .mdtable import (
        FileRow,
        EventRow,
        FieldRow,
        ParamRow,
        ModuleRow,
        TypeDefRow,
        TypeRefRow,
        AssemblyRow,
        PropertyRow,
        TypeSpecRow,
        MemberRefRow,
        MethodDefRow,
        ModuleRefRow,
        AssemblyRefRow,
        DeclSecurityRow,
        ExportedTypeRow,
        GenericParamRow,
        InterfaceImplRow,
        StandAloneSigRow,
        ManifestResourceRow,
        GenericParamConstraintRow,
    )


class TypeDefOrRef(CodedIndex[Union["TypeDefRow", "TypeRefRow", "TypeSpecRow"]]):
    tag_bits = 2
    table_names = ("TypeDef", "TypeRef", "TypeSpec")
    table_numbers = (2, 1, 27)


class HasConstant(CodedIndex[Union["FieldRow", "ParamRow", "PropertyRow"]]):
    tag_bits = 2
    table_names = ("Field", "Param", "Property")
    table_numbers = (4, 8, 23)


class HasCustomAttribute(
    CodedIndex[
        Union[
            "MethodDefRow",
            "FieldRow",
            "TypeRefRow",
            "TypeDefRow",
            "ParamRow",
            "InterfaceImplRow",
            "MemberRefRow",
            "ModuleRow",
            "DeclSecurityRow",
            "PropertyRow",
            "EventRow",
            "StandAloneSigRow",
            "ModuleRefRow",
            "TypeSpecRow",
            "AssemblyRow",
            "AssemblyRefRow",
            "FileRow",
            "ExportedTypeRow",
            "ManifestResourceRow",
            "GenericParamRow",
            "GenericParamConstraintRow",
        ]
    ]
):
    tag_bits = 5
    # TODO this may be an incomplete list
    table_names = (
        "MethodDef",
        "Field",
        "TypeRef",
        "TypeDef",
        "Param",
        "InterfaceImpl",
        "MemberRef",
        "Module",
        "DeclSecurity",
        "Property",
        "Event",
        "StandAloneSig",
        "ModuleRef",
        "TypeSpec",
        "Assembly",
        "AssemblyRef",
        "File",
        "ExportedType",
        "ManifestResource",
        "GenericParam",
        "GenericParamConstraint",
    )


class HasFieldMarshall(CodedIndex[Union["FieldRow", "ParamRow"]]):
    tag_bits = 1
    table_names = ("Field", "Param")


class HasDeclSecurity(CodedIndex[Union["TypeDefRow", "MethodDefRow", "AssemblyRow"]]):
    tag_bits = 2
    table_names = ("TypeDef", "MethodDef", "Assembly")


class MemberRefParent(CodedIndex[Union["TypeDefRow", "TypeRefRow", "ModuleRefRow", "MethodDefRow", "TypeSpecRow"]]):
    tag_bits = 3
    table_names = ("TypeDef", "TypeRef", "ModuleRef", "MethodDef", "TypeSpec")


class HasSemantics(CodedIndex[Union["EventRow", "PropertyRow"]]):
    tag_bits = 1
    table_names = ("Event", "Property")


class MethodDefOrRef(CodedIndex[Union["MethodDefRow", "MemberRefRow"]]):
    tag_bits = 1
    table_names = ("MethodDef", "MemberRef")


class MemberForwarded(CodedIndex[Union["FieldRow", "MethodDefRow"]]):
    tag_bits = 1
    table_names = ("Field", "MethodDef")


class Implementation(CodedIndex[Union["FileRow", "AssemblyRefRow", "ExportedTypeRow"]]):
    tag_bits = 2
    table_names = ("File", "AssemblyRef", "ExportedType")


class CustomAttributeType(CodedIndex[Union["MethodDefRow", "MemberRefRow"]]):
    tag_bits = 3
    table_names = ("Unused", "Unused", "MethodDef", "MemberRef", "Unused")


class ResolutionScope(CodedIndex[Union["ModuleRow", "ModuleRefRow", "AssemblyRefRow", "TypeRefRow"]]):
    tag_bits = 2
    table_names = ("Module", "ModuleRef", "AssemblyRef", "TypeRef")


class TypeOrMethodDef(CodedIndex[Union["TypeDefRow", "MethodDefRow"]]):
    tag_bits = 1
    table_names = ("TypeDef", "MethodDef")



import fixtures

import dnfile


def test_duplicate_stream():
    path = fixtures.DATA / "invalid-streams" / "duplicate-stream.exe"

    dn = dnfile.dnPE(path)

    assert b"#US" in dn.net.metadata.streams
    assert dn.net.user_strings.get(1).value == "BBBBBBBB"


def test_unknown_stream():
    path = fixtures.DATA / "invalid-streams" / "unknown-stream.exe"

    dn = dnfile.dnPE(path)

    assert b"#ZZ" in dn.net.metadata.streams


def test_invalid_stream_name():
    path = fixtures.DATA / "invalid-streams" / "invalid-stream-name.exe"

    dn = dnfile.dnPE(path)

    assert b"#\x90\x90" in dn.net.metadata.streams



# -*- coding: utf-8 -*-

import copy as _copymod
import logging
import functools as _functools
from typing import List, Tuple, TypeVar, Optional, cast
from collections import deque

logger = logging.getLogger(__name__)


# lru_cache with a shallow copy of the objects returned (list, dicts, ..)
# we don't use deepcopy as it's _really_ slow and the data we retrieved using this is enough with copy.copy
# taken from https://stackoverflow.com/questions/54909357/how-to-get-functools-lru-cache-to-return-new-instances
def lru_cache(maxsize=128, typed=False, copy=False):
    if not copy:
        return _functools.lru_cache(maxsize, typed)

    def decorator(f):
        cached_func = _functools.lru_cache(maxsize, typed)(f)

        @_functools.wraps(f)
        def wrapper(*args, **kwargs):
            # return _copymod.deepcopy(cached_func(*args, **kwargs))
            return _copymod.copy(cached_func(*args, **kwargs))

        return wrapper

    return decorator


def read_compressed_int(data) -> Optional[Tuple[int, int]]:
    """
    Given bytes, read a compressed integer per
    spec ECMA-335 II.23.2 Blobs and signatures.
    Returns tuple: value, number of bytes read or None on error.
    """
    if not data:
        return None
    if data[0] & 0x80 == 0:
        # values 0x00 to 0x7f
        return data[0], 1
    elif len(data) >= 2 and data[0] & 0x40 == 0:
        # values 0x80 to 0x3fff
        value = (data[0] & 0x7F) << 8
        value |= data[1]
        return value, 2
    elif len(data) >= 4 and data[0] & 0x20 == 0:
        # values 0x4000 to 0x1fffffff
        value = (data[0] & 0x3F) << 24
        value |= data[1] << 16
        value |= data[2] << 8
        value |= data[3]
        return value, 4
    else:
        logger.warning("invalid compressed int: leading byte: 0x%02x", data[0])
        return None


def compress_int(i: int) -> Optional[bytes]:
    """
    Given integer, return bytes representing a compressed integer per
    spec ECMA-335 II.23.2 Blobs and signatures.
    """

    if not isinstance(i, int):
        raise TypeError(f"expected int, given {type(i)}")
    if i < 0:
        raise ValueError(f"expected positive int, given {i}")

    if i <= 0x7f:
        return int.to_bytes(i, length=1, byteorder="big")
    elif i <= 0x3fff:
        b = int.to_bytes(i, length=2, byteorder="big")
        return bytes((0x80 | b[0], b[1]))
    elif i <= 0x1fffffff:
        b = int.to_bytes(i, length=4, byteorder="big")
        return bytes((0x80 | 0x40 | b[0], b[1], b[2], b[3]))
    else:
        logger.warning(f"invalid int {i}, max value 0x1fffffff")
        return None


def two_way_dict(pairs):
    return dict([(e[1], e[0]) for e in pairs] + pairs)


def num_bytes_to_struct_char(n: int) -> Optional[str]:
    """
    Given number of bytes, return the struct char that can hold those bytes.
    Returns None on invalid value.

    For example,
        2 = H
        4 = I
    """
    if n > 8:
        logger.warning("invalid format specifier: %d > 8", n)
        return None
    elif n > 4:
        return "Q"
    elif n > 2:
        return "I"
    elif n > 1:
        return "H"
    elif n == 1:
        return "B"
    else:
        logger.warning("invalid format specifier: %d", n)
        return None


_ListType = TypeVar("_ListType")


class LazyList(List[_ListType]):
    """A List that transparently initializes each item as it is accessed.

    Note that calling `str` or `repr` with an instance of this class will force
    all items to be initialized.
    """

    def __init__(self, eval_func, initial_size: int):
        """Create a lazily initialized list of the specified size.

        `eval_func` is called with the index and value (or slice and list of values)
        for items retrieved through `__getitem__` *every time they are accessed*.
        """
        self.eval_func = eval_func
        super().__init__(cast(List[_ListType], [None] * initial_size))

    def __getitem__(self, __key):
        __value = super().__getitem__(__key)
        new = self.eval_func(__key, __value)
        if __value != new:
            super().__setitem__(__key, new)
        return new

    def __iter__(self):
        i = 0  # `eval_func` requires an index
        for v in super().__iter__():
            new = self.eval_func(i, v)
            if v != new:
                super().__setitem__(i, new)
            yield new
            i += 1

    def eval_all(self):
        """Ensure all items in the list are initialized by iterating through them."""
        deque(self, maxlen=0)  # Fully consumes iterator without overhead (https://stackoverflow.com/a/50938015)

    def truncate(self, length: int):
        """Truncate the list without evaluating any items."""
        keep = super().__getitem__(slice(0, length))
        self.clear()
        self.extend(keep)

    def __repr__(self) -> str:
        self.eval_all()
        return super().__repr__()



import pytest
import fixtures

import dnfile


def test_unpaired_surrogate():
    path = fixtures.DATA / "invalid-strings" / "unpaired-surrogate.exe"

    dn = dnfile.dnPE(path)

    assert dn.net is not None
    assert dn.net.metadata is not None
    assert dn.net.user_strings is not None

    assert b"#US" in dn.net.metadata.streams
    item = dn.net.user_strings.get(1)
    assert item is not None
    assert item.flag == 0x01
    assert item.value_bytes() == b"\xD0\xDD"
    assert item.value is None


def test_raw_binary():
    path = fixtures.DATA / "invalid-strings" / "raw-binary.exe"

    dn = dnfile.dnPE(path)

    assert dn.net is not None
    assert dn.net.metadata is not None
    assert dn.net.user_strings is not None

    # short MZ header
    assert b"#US" in dn.net.metadata.streams
    s = dn.net.user_strings.get(1)
    assert s is not None
    # somehow this is valid utf-16
    assert s.value == b"\x4D\x5A\x90\x00".decode("utf-16")
    assert s.value_bytes() == b"\x4D\x5A\x90\x00"


def test_string_decoder():
    path = fixtures.DATA / "invalid-strings" / "string-decoder.exe"

    dn = dnfile.dnPE(path)

    assert dn.net is not None
    assert dn.net.metadata is not None
    assert dn.net.user_strings is not None

    # "Hello World" ^ 0xFF
    assert b"#US" in dn.net.metadata.streams
    item = dn.net.user_strings.get(1)
    assert item is not None
    assert item.raw_data == b"\x17\xb7\xff\x9a\xff\x93\xff\x93\xff\x90\xff\xdf\xff\xa8\xff\x90\xff\x8d\xff\x93\xff\x9b\xff\x01"
    assert item.flag == 0x01
    assert item.value_bytes() ==  b"\xb7\xff\x9a\xff\x93\xff\x93\xff\x90\xff\xdf\xff\xa8\xff\x90\xff\x8d\xff\x93\xff\x9b\xff"

    # somehow this is valid utf-16
    s = dn.net.user_strings.get(1)
    assert s is not None
